import weakref

from defcon.objects.base import BaseObject
from fontTools.ttLib import TTFont
from ufo2ft.featureCompiler import FeatureCompiler

from defconQt.controls.glyphContextView import GlyphRecord

try:
    from harfbuzz import HARFBUZZ as HB
    import harfbuzz as hb
except ImportError:
    pass

CH_GID_PREFIX = 0x80000000

# ---------
# Factories
# ---------


def _layoutEngineOTLTablesRepresentationFactory(layoutEngine):
    font = layoutEngine.font
    ret = dict()
    glyphOrder = sorted(font.keys())
    if font.features.text:
        otf = TTFont()
        otf.setGlyphOrder(glyphOrder)
        # compile with feaLib + markWriter/kernWriter
        try:
            compiler = FeatureCompiler(font, otf)
            compiler.postProcess = lambda: None
            compiler.compile()
            for name in ("GDEF", "GSUB", "GPOS"):
                if name in otf:
                    table = otf[name].compile(otf)
                    value = hb.Blob.create_for_array(table, HB.MEMORY_MODE_READONLY)
                    ret[name] = value
        except Exception:
            # TODO: handle this in the UI
            import traceback

            print(traceback.format_exc(5))
            # discard tables from incompletely parsed feature text
            ret = dict()
    return ret, glyphOrder


# harfbuzz


def _get_nominal_glyph(font, engine, ch, user_data):
    if ch >= CH_GID_PREFIX:
        return ch - CH_GID_PREFIX

    ufo = engine.font
    glyphName = ufo.unicodeData.glyphNameForUnicode(ch)
    if glyphName is not None:
        try:
            return engine.glyphOrder.index(glyphName)
        except IndexError:
            pass
    return 0


def _get_glyph_h_advance(font, engine, gid, user_data):
    ufo = engine.font
    glyph = ufo[engine.glyphOrder[gid]]
    return glyph.width * font.scale[0] / font.face.upem


def _get_glyph_name_func(font, engine, gid, user_data):
    return engine.glyphOrder[gid]


def _spitLayoutTable(face, tag, layoutTables):
    name = hb.tag_to_string(tag)
    return layoutTables.get(name)


# TODO: what if the source font does not have .notdef, is there a
# fallback gid?


class LayoutEngine(BaseObject):
    changeNotificationName = "LayoutEngine.Changed"
    representationFactories = {
        "TruFont.layoutEngine.tables": dict(
            factory=_layoutEngineOTLTablesRepresentationFactory,
            destructiveNotifications=("LayoutEngine._DestroyCachedTables"),
        )
    }

    def __init__(self, font):
        self._needsInternalUpdate = True
        self._font = weakref.ref(font)
        self._fontFeatures = dict()
        self._glyphOrder = []
        self._hbFont = None
        super().__init__()
        self.beginSelfNotificationObservation()

    @property
    def font(self):
        if self._font is not None:
            return self._font()
        return None

    @property
    def glyphOrder(self):
        return self._glyphOrder

    # --------------
    # Engine Updates
    # --------------

    def _updateEngine(self):
        if not self._needsInternalUpdate:
            return
        ufo = self.font
        layoutTables, self._glyphOrder = self.getRepresentation(
            "TruFont.layoutEngine.tables"
        )

        face = hb.Face.create_for_tables(_spitLayoutTable, layoutTables, None, False)
        font = hb.Font.create(face)
        face.upem = upem = ufo.info.unitsPerEm
        font.scale = (upem, upem)

        funcs = hb.FontFuncs.create(False)
        funcs.set_nominal_glyph_func(_get_nominal_glyph, None, None)
        funcs.set_glyph_h_advance_func(_get_glyph_h_advance, None, None)
        # TODO: vertical advance
        funcs.set_glyph_name_func(_get_glyph_name_func, None, None)
        font.set_funcs(funcs, self, None)

        self._hbFont = font
        self._needsInternalUpdate = False

    # -------------
    # Notifications
    # -------------

    def beginSelfNotificationObservation(self):
        super().beginSelfNotificationObservation()
        self.beginSelfLayersObservation()
        self.beginSelfLayerObservation()
        self.beginSelfFeaturesObservation()

    def endSelfNotificationObservation(self):
        self.endSelfLayersObservation()
        self.endSelfLayerObservation()
        self.endSelfFeaturesObservation()
        super().endSelfNotificationObservation()
        self._font = None

    # default layer changed (changes cmap)

    def beginSelfLayersObservation(self):
        layers = self.font.layers
        layers.addObserver(
            observer=self,
            methodName="_layerSetDefaultLayerWillChange",
            notification="LayerSet.DefaultLayerWillChange",
        )
        layers.addObserver(
            observer=self,
            methodName="_layerSetDefaultLayerChanged",
            notification="LayerSet.DefaultLayerChanged",
        )

    def endSelfLayersObservation(self):
        layers = self.font.layers
        layers.removeObserver(
            observer=self, notification="LayerSet.DefaultLayerWillChange"
        )
        layers.removeObserver(
            observer=self, notification="LayerSet.DefaultLayerChanged"
        )

    def _layerSetDefaultLayerWillChange(self, notification):
        self.endSelfLayerObservation()

    def _layerSetDefaultLayerChanged(self, notification):
        self.beginLayerObservation()
        self._postNeedsUpdateNotification()

    # cmap change

    def beginSelfLayerObservation(self):
        layer = self.font.layers.defaultLayer
        layer.addObserver(
            observer=self,
            methodName="_layerGlyphNameChanged",
            notification="Layer.GlyphNameChanged",
        )
        layer.addObserver(
            observer=self,
            methodName="_layerGlyphUnicodesChanged",
            notification="Layer.GlyphUnicodesChanged",
        )

    def endSelfLayerObservation(self):
        layer = self.font.layers.defaultLayer
        layer.removeObserver(observer=self, notification="Layer.GlyphNameChanged")
        layer.removeObserver(observer=self, notification="Layer.GlyphUnicodesChanged")

    def _layerGlyphNameChanged(self, notification):
        self._postNeedsUpdateNotification()

    def _layerGlyphUnicodesChanged(self, notification):
        self._postNeedsUpdateNotification()

    # feature text change

    def beginSelfFeaturesObservation(self):
        features = self.font.features
        features.addObserver(
            observer=self,
            methodName="_featuresTextChanged",
            notification="Features.TextChanged",
        )

    def endSelfFeaturesObservation(self):
        features = self.font.features
        features.removeObserver(observer=self, notification="Features.TextChanged")

    def _featuresTextChanged(self, notification):
        self._destroyCachedTables()
        self._postNeedsUpdateNotification()

    # posting

    def _destroyCachedTables(self):
        self.postNotification("LayoutEngine._DestroyCachedTables")

    def _postNeedsUpdateNotification(self):
        self._needsInternalUpdate = True
        self.postNotification(self.changeNotificationName)

    # ----------
    # Engine API
    # ----------

    def process(self, text):
        if not text:
            return []
        if self._needsInternalUpdate:
            self._updateEngine()

        # TODO: reuse buffer?
        buf = hb.Buffer.create()
        if isinstance(text, list):
            font = self.font
            unicodes = []
            for name in text:
                uni = font.unicodeData.unicodeForGlyphName(name)
                if uni is None:
                    # TODO: use a dict instead?
                    gid = self._glyphOrder.index(name)
                    uni = CH_GID_PREFIX + gid
                unicodes.append(uni)
            buf.add_codepoints(unicodes, len(unicodes), 0, len(unicodes))
        else:
            buf.add_str(text)
        buf.guess_segment_properties()
        hb.shape(self._hbFont, buf, list(self._fontFeatures.values()))

        glyphRecords = []
        for info, pos in zip(buf.glyph_infos, buf.glyph_positions):
            glyphName = self._glyphOrder[info.codepoint]
            if glyphName not in self.font:
                continue
            record = GlyphRecord()
            record.glyph = self.font[glyphName]
            record.cluster = info.cluster
            record.xOffset = pos.x_offset
            record.yOffset = pos.y_offset
            record.xAdvance = pos.x_advance
            record.yAdvance = pos.y_advance
            glyphRecords.append(record)
        del buf
        return glyphRecords

    def getScriptList(self):
        self._updateEngine()
        # XXX: for now we only list GSUB
        tags = self._hbFont.face.ot_layout.table_get_script_tags(
            hb.tag_from_string("GSUB")
        )
        return [hb.tag_to_string(t) for t in tags]

    def getLanguageList(self):
        self._updateEngine()
        # XXX: for now we only list GSUB, default script
        tags = self._hbFont.face.ot_layout.script_get_language_tags(
            hb.tag_from_string("GSUB"), 0
        )
        return [hb.tag_to_string(t) for t in tags]

    def getFeatureList(self):
        self._updateEngine()
        # XXX: for now we only list GSUB, default script, default language
        tags = self._hbFont.face.ot_layout.language_get_feature_tags(
            hb.tag_from_string("GSUB"), 0, 0xFFFF
        )
        return [hb.tag_to_string(t) for t in tags]

    def getFeatureState(self, name):
        ret = self._fontFeatures.get(name)
        if ret is not None:
            return ret.value
        return ret

    def setFeatureState(self, name, state):
        if state is None:
            if name in self._fontFeatures:
                del self._fontFeatures[name]
            return
        #
        if name in self._fontFeatures:
            feature = self._fontFeatures[name]
        else:
            feature = hb.Feature.from_string(name)
        feature.value = state
        # TODO: start, end
        self._fontFeatures[name] = feature



import os

from PyQt5.QtCore import QRegularExpression, QSize
from PyQt5.QtGui import QRegularExpressionValidator
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QFormLayout,
    QFrame,
    QLabel,
    QLineEdit,
    QSizePolicy,
    QVBoxLayout,
)

from defconQt.controls.listView import ListView
from trufont.controls.folderComboBox import FolderComboBox
from trufont.objects.extension import TExtension


def VersionValidator(parent):
    validator = QRegularExpressionValidator(parent)
    validator.setRegularExpression(QRegularExpression("([0-9]+\\.[0-9]+\\.[0-9]+)?"))
    return validator


def HLine(parent):
    widget = QFrame(parent)
    widget.setFrameShape(QFrame.HLine)
    widget.setFrameShadow(QFrame.Sunken)
    return widget


class ExtensionBuilderWindow(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("Extension Builder"))

        layout = QFormLayout(self)
        layout.setFieldGrowthPolicy(QFormLayout.ExpandingFieldsGrow)

        self.nameEdit = QLineEdit(self)
        self.versionEdit = QLineEdit(self)
        self.versionEdit.setValidator(VersionValidator(self))
        self.developerEdit = QLineEdit(self)
        self.developerURLEdit = QLineEdit(self)
        layout.addRow(self.tr("Name:"), self.nameEdit)
        layout.addRow(self.tr("Version:"), self.versionEdit)
        layout.addRow(self.tr("Developer:"), self.developerEdit)
        layout.addRow(self.tr("Developer URL:"), self.developerURLEdit)
        layout.addRow(HLine(self))

        self.resourcesRootBox = FolderComboBox(self)
        self.resourcesRootBox.setSizePolicy(
            QSizePolicy.Expanding, QSizePolicy.Preferred
        )
        self.scriptRootBox = FolderComboBox(self)
        self.scriptRootBox.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        self.scriptRootBox.currentFolderModified.connect(self.updateView)
        layout.addRow(self.tr("Resources root:"), self.resourcesRootBox)
        layout.addRow(self.tr("Script root:"), self.scriptRootBox)

        self.launchAtStartupBox = QCheckBox(self.tr("Launch At Startup"), self)
        self.mainScriptDrop = QComboBox(self)
        addScriptsLabel = QLabel(self.tr("Add script to main menu:"), self)
        self.addScriptsView = ListView(self)
        self.addScriptsView.setList([["", "", "", ""]])
        self.addScriptsView.setHeaderLabels(["", "Script", "Menu name", "Shortcut"])
        scriptLayout = QVBoxLayout()
        scriptLayout.addWidget(self.launchAtStartupBox)
        scriptLayout.addWidget(self.mainScriptDrop)
        scriptLayout.addWidget(addScriptsLabel)
        scriptLayout.addWidget(self.addScriptsView)
        layout.addRow("", scriptLayout)
        layout.addRow(HLine(self))

        self.tfVersionEdit = QLineEdit(self)
        self.tfVersionEdit.setValidator(VersionValidator(self))
        layout.addRow(self.tr("Requires TruFont:"), self.tfVersionEdit)

        buttonBox = QDialogButtonBox(
            QDialogButtonBox.Save | QDialogButtonBox.Close, self
        )
        buttonBox.accepted.connect(self.saveFile)
        buttonBox.rejected.connect(self.close)
        layout.addRow(buttonBox)

        self.setLayout(layout)
        self.updateView()

    def saveFile(self):
        e = TExtension()
        e.name = self.nameEdit.text() or None
        e.version = self.versionEdit.text() or None
        e.developer = self.developerEdit.text() or None
        e.developerURL = self.developerURLEdit.text() or None

        e.resourcesPath = self.resourcesRootBox.currentFolder()
        e.libPath = self.scriptRootBox.currentFolder()

        e.launchAtStartup = self.launchAtStartupBox.isChecked()
        e.mainScript = self.mainScriptDrop.currentText()
        # Note. for now we always do a list.
        addToMenu = []
        for ok, path, name, shortcut in self.addScriptsView.list():
            if not ok:
                continue
            data = dict(path=path, name=name, shortcut=shortcut)
            for k, v in list(data.items()):
                if v is None:
                    del data[k]
            addToMenu.append(data)
        if addToMenu:
            e.addToMenu = addToMenu
        e.tfVersion = self.tfVersionEdit.text() or None

        # TODO: switch to directory on platforms that need it
        dialog = QFileDialog(
            self, self.tr("Save File"), None, "TruFont Extension (*.tfExt)"
        )
        dialog.setAcceptMode(QFileDialog.AcceptSave)
        ok = dialog.exec_()
        if ok:
            path = dialog.selectedFiles()[0]
            e.save(path)

    def updateView(self):
        path = self.scriptRootBox.currentFolder()
        widgets = (self.launchAtStartupBox, self.mainScriptDrop, self.addScriptsView)
        for widget in widgets:
            widget.setEnabled(path is not None)
        if path is None:
            return
        elements = []
        for root, _, files in os.walk(path):
            for file in files:
                if os.path.splitext(file)[1] != ".py":
                    continue
                name = os.path.join(root[len(path) + 1 :], file)
                self.mainScriptDrop.addItem(name)
                elements.append([False, name, None, ""])
        self.addScriptsView.setList(elements)
        # TODO: should be done in the widget
        model = self.addScriptsView.model()
        for i in range(len(elements)):
            self.addScriptsView.openPersistentEditor(model.index(i, 0))

    # ----------
    # Qt methods
    # ----------

    def sizeHint(self):
        return QSize(580, 480)



from PyQt5.QtCore import QEvent, QLocale, Qt
from PyQt5.QtGui import QDoubleValidator
from PyQt5.QtWidgets import (
    QDialog,
    QDialogButtonBox,
    QGridLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QRadioButton,
)


class FindDialog(QDialog):
    alphabetical = [dict(type="alphabetical", allowPseudoUnicode=True)]

    def __init__(self, currentGlyph, parent=None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModal)
        self.setWindowTitle(self.tr("Find…"))
        self._sortedGlyphNames = currentGlyph.font.unicodeData.sortGlyphNames(
            currentGlyph.layer.keys(), self.alphabetical
        )

        layout = QGridLayout(self)
        self.glyphLabel = QLabel(self.tr("Glyph:"), self)
        self.glyphEdit = QLineEdit(self)
        self.glyphEdit.textChanged.connect(self.updateGlyphList)
        self.glyphEdit.event = self.lineEvent
        self.glyphEdit.keyPressEvent = self.lineKeyPressEvent

        self.beginsWithBox = QRadioButton(self.tr("Begins with"), self)
        self.containsBox = QRadioButton(self.tr("Contains"), self)
        self.beginsWithBox.setChecked(True)
        self.beginsWithBox.toggled.connect(self.updateGlyphList)

        self.glyphList = QListWidget(self)
        self.glyphList.itemDoubleClicked.connect(self.accept)

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        line = 0
        layout.addWidget(self.glyphLabel, line, 0, 1, 2)
        layout.addWidget(self.glyphEdit, line, 2, 1, 4)
        line += 1
        layout.addWidget(self.beginsWithBox, line, 0, 1, 3)
        layout.addWidget(self.containsBox, line, 3, 1, 3)
        line += 1
        layout.addWidget(self.glyphList, line, 0, 1, 6)
        line += 1
        layout.addWidget(buttonBox, line, 0, 1, 6)
        self.setLayout(layout)
        self.updateGlyphList()

    def lineEvent(self, event):
        if event.type() == QEvent.KeyPress and event.key() == Qt.Key_Tab:
            if self.beginsWithBox.isChecked():
                self.containsBox.toggle()
            else:
                self.beginsWithBox.toggle()
            return True
        else:
            return QLineEdit.event(self.glyphEdit, event)

    def lineKeyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key_Up or key == Qt.Key_Down:
            self.glyphList.keyPressEvent(event)
        else:
            QLineEdit.keyPressEvent(self.glyphEdit, event)

    def updateGlyphList(self):
        beginsWith = self.beginsWithBox.isChecked()
        self.glyphList.clear()
        if not self.glyphEdit.isModified():
            self.glyphList.addItems(self._sortedGlyphNames)
        else:
            text = self.glyphEdit.text()
            if beginsWith:
                glyphs = [
                    glyphName
                    for glyphName in self._sortedGlyphNames
                    if glyphName and glyphName.startswith(text)
                ]
            else:
                glyphs = [
                    glyphName
                    for glyphName in self._sortedGlyphNames
                    if glyphName and text in glyphName
                ]
            self.glyphList.addItems(glyphs)
        self.glyphList.setCurrentRow(0)

    @classmethod
    def getNewGlyph(cls, parent, currentGlyph):
        dialog = cls(currentGlyph, parent)
        result = dialog.exec_()
        currentItem = dialog.glyphList.currentItem()
        newGlyph = None
        if currentItem is not None:
            newGlyphName = currentItem.text()
            if newGlyphName in currentGlyph.layer:
                newGlyph = currentGlyph.layer[newGlyphName]
        return (newGlyph, result)


class AddComponentDialog(FindDialog):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setWindowTitle(self.tr("Add component…"))
        self._sortedGlyphNames.remove(args[0].name)
        self.updateGlyphList()


class LayerActionsDialog(QDialog):
    def __init__(self, currentGlyph, parent=None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModal)
        self.setWindowTitle(self.tr("Layer actions…"))
        self._workableLayers = []
        for layer in currentGlyph.layerSet:
            if layer != currentGlyph.layer:
                self._workableLayers.append(layer)

        copyBox = QRadioButton(self.tr("Copy"), self)
        moveBox = QRadioButton(self.tr("Move"), self)
        swapBox = QRadioButton(self.tr("Swap"), self)
        self.otherCheckBoxes = (moveBox, swapBox)
        copyBox.setChecked(True)

        self.layersList = QListWidget(self)
        self.layersList.addItems(layer.name for layer in self._workableLayers)
        if self.layersList.count():
            self.layersList.setCurrentRow(0)
        self.layersList.itemDoubleClicked.connect(self.accept)

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        layout = QGridLayout(self)
        line = 0
        layout.addWidget(copyBox, line, 0, 1, 2)
        layout.addWidget(moveBox, line, 2, 1, 2)
        layout.addWidget(swapBox, line, 4, 1, 2)
        line += 1
        layout.addWidget(self.layersList, line, 0, 1, 6)
        line += 1
        layout.addWidget(buttonBox, line, 0, 1, 6)
        self.setLayout(layout)

    @classmethod
    def getLayerAndAction(cls, parent, currentGlyph):
        dialog = cls(currentGlyph, parent)
        result = dialog.exec_()
        currentItem = dialog.layersList.currentItem()
        newLayer = None
        if currentItem is not None:
            newLayerName = currentItem.text()
            for layer in dialog._workableLayers:
                if layer.name == newLayerName:
                    newLayer = layer
        action = "Copy"
        for checkBox in dialog.otherCheckBoxes:
            if checkBox.isChecked():
                action = checkBox.text()
        return (newLayer, action, result)


class EditDialog(QDialog):
    def __init__(self, parent=None, item=None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModal)
        self.setWindowTitle(self.tr("Edit…"))

        nameLabel = QLabel(self.tr("Name:"), self)
        self.nameEdit = QLineEdit(self)
        self.nameEdit.setFocus(Qt.OtherFocusReason)

        validator = QDoubleValidator(self)
        validator.setLocale(QLocale.c())
        xLabel = QLabel(self.tr("X:"), self)
        self.xEdit = QLineEdit(self)
        self.xEdit.setValidator(validator)
        yLabel = QLabel(self.tr("Y:"), self)
        self.yEdit = QLineEdit(self)
        self.yEdit.setValidator(validator)

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        layout = QGridLayout(self)
        line = 0
        layout.addWidget(nameLabel, line, 0)
        layout.addWidget(self.nameEdit, line, 1, 1, 3)
        line += 1
        layout.addWidget(xLabel, line, 0)
        layout.addWidget(self.xEdit, line, 1)
        layout.addWidget(yLabel, line, 2)
        layout.addWidget(self.yEdit, line, 3)
        line += 1
        layout.addWidget(buttonBox, line, 3)
        self.setLayout(layout)

    @classmethod
    def getNewProperties(cls, parent, item):
        dialog = cls(parent, item)
        dialog.nameEdit.setText(item.name)
        dialog.nameEdit.selectAll()
        dialog.xEdit.setText(str(item.x))
        dialog.xEdit.selectAll()
        dialog.yEdit.setText(str(item.y))
        dialog.yEdit.selectAll()
        result = dialog.exec_()
        name = dialog.nameEdit.text()
        x = float(dialog.xEdit.text())
        y = float(dialog.yEdit.text())
        return (name, x, y, result)



from PyQt5.QtGui import QColor, QFont, QPainter
from PyQt5.QtWidgets import QLabel, QSizePolicy, QVBoxLayout, QWidget

from trufont.tools import platformSpecific


class GroupBox(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setSizePolicy(QSizePolicy.Minimum, QSizePolicy.Preferred)
        pointSize, delta = 8, platformSpecific.fontSizeDelta()
        self.setStyleSheet(
            "QLabel {{ color: #505050; font-size: {}pt }}".format(
                pointSize + 1 + 3 * delta
            )
        )

        self.titleLabel = QLabel(self)
        font = self.titleLabel.font()
        font.setCapitalization(QFont.AllUppercase)
        font.setLetterSpacing(QFont.AbsoluteSpacing, 1)
        font.setPointSize(pointSize + delta)
        self.titleLabel.setFont(font)
        self.titleLabel.setStyleSheet(
            "color: #787878; font-size: {}pt".format(pointSize + delta)
        )

        layout = QVBoxLayout(self)
        layout.addWidget(self.titleLabel)
        layout.setContentsMargins(12, 10, 12, 16)
        layout.setSpacing(16)

    def setChildLayout(self, layout):
        layout.setContentsMargins(0, 0, 0, 0)
        self.layout().addLayout(layout)

    def title(self):
        return self.titleLabel.text()

    def setTitle(self, text):
        self.titleLabel.setText(text)

    # ----------
    # Qt methods
    # ----------

    def paintEvent(self, event):
        painter = QPainter(self)
        painter.fillRect(event.rect(), QColor(240, 240, 240))



import os

from defcon.tools.notifications import NotificationCenter
from PyQt5.QtCore import QEvent, QStandardPaths, Qt, QUrl
from PyQt5.QtGui import QDesktopServices
from PyQt5.QtWidgets import QAction, QApplication, QFileDialog

from trufont.controls.aboutDialog import AboutDialog
from trufont.drawingTools.knifeTool import KnifeTool
from trufont.drawingTools.penTool import PenTool
from trufont.drawingTools.rulerTool import RulerTool
from trufont.drawingTools.selectionTool import SelectionTool
from trufont.drawingTools.shapesTool import ShapesTool
from trufont.drawingTools.textTool import TextTool
from trufont.objects import settings
from trufont.objects.defcon import TFont
from trufont.objects.extension import TExtension
from trufont.objects.menu import MAX_RECENT_FILES, Entries, MenuBar, globalMenuBar
from trufont.tools import errorReports, glyphList, platformSpecific
from trufont.windows.extensionBuilderWindow import ExtensionBuilderWindow
from trufont.windows.fontWindow import FontWindow
from trufont.windows.scriptingWindow import ScriptingWindow
from trufont.windows.settingsWindow import SettingsWindow


class Application(QApplication):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self._currentGlyph = None
        self._currentFontWindow = None
        self._launched = False
        self._drawingTools = [
            SelectionTool,
            PenTool,
            KnifeTool,
            RulerTool,
            ShapesTool,
            TextTool,
        ]
        self._extensions = []
        self.dispatcher = NotificationCenter()
        self.dispatcher.addObserver(self, "_fontWindowClosed", "fontWillClose")
        self.focusWindowChanged.connect(self._focusWindowChanged)
        self.GL2UV = None
        self.outputWindow = None

    # --------------
    # Event handling
    # --------------

    def _focusWindowChanged(self):
        # update menu bar
        self.updateMenuBar()
        # update main window
        window = self.activeWindow()
        if window is None:
            return
        while True:
            parent = window.parent()
            if parent is None:
                break
            window = parent
        if isinstance(window, FontWindow):
            self.setCurrentFontWindow(window)

    def _fontWindowClosed(self, notification):
        font = notification.data["font"]
        # cleanup CurrentFont/CurrentGlyph when closing the corresponding
        # window
        if self._currentFontWindow is not None:
            if self._currentFontWindow.font == font:
                self.setCurrentFontWindow(None)
        if self._currentGlyph is not None:
            if self._currentGlyph.font == font:
                self.setCurrentGlyph(None)

    def event(self, event):
        eventType = event.type()
        # respond to OSX open events
        if eventType == QEvent.FileOpen:
            filePath = event.file()
            self.openFile(filePath)
            return True
        elif eventType == QEvent.ApplicationStateChange:
            applicationState = self.applicationState()
            if applicationState == Qt.ApplicationActive:
                if not self._launched:
                    notification = "applicationLaunched"
                    self.loadGlyphList()
                    self._launched = True
                else:
                    notification = "applicationActivated"
                    # XXX: do it
                    # self.lookupExternalChanges()
                self.postNotification(notification)
            elif applicationState == Qt.ApplicationInactive:
                self.postNotification("applicationWillIdle")
        return super().event(event)

    def postNotification(self, notification, data=None):
        dispatcher = self.dispatcher
        dispatcher.postNotification(
            notification=notification, observable=self, data=data
        )

    # ---------------
    # File management
    # ---------------

    def loadGlyphList(self):
        glyphListPath = settings.glyphListPath()
        if glyphListPath and os.path.exists(glyphListPath):
            try:
                glyphList_ = glyphList.parseGlyphList(glyphListPath)
            except Exception as e:
                msg = self.tr(
                    "The glyph list at {0} cannot " "be parsed and will be dropped."
                ).format(glyphListPath)
                errorReports.showWarningException(e, msg)
                settings.removeGlyphListPath()
            else:
                self.GL2UV = glyphList_

    def lookupExternalChanges(self):
        for font in self.allFonts():
            if not font.path:
                continue
            changed = font.testForExternalChanges()
            for attr in ("info", "kerning", "groups", "features", "lib"):
                if changed[attr]:
                    data = dict(font=font)
                    self.postNotification("fontChangedExternally", data)
                    return
            # XXX: do more

    # -----------------
    # Window management
    # -----------------

    def currentFontWindow(self):
        return self._currentFontWindow

    def setCurrentFontWindow(self, fontWindow):
        if fontWindow == self._currentFontWindow:
            return
        self._currentFontWindow = fontWindow
        self.postNotification("currentFontChanged")

    # --------
    # Menu Bar
    # --------

    def fetchMenuBar(self, window=None):
        if platformSpecific.useGlobalMenuBar():
            try:
                self._menuBar
            except Exception:
                self._menuBar = globalMenuBar()
            self._menuBar.resetState()
            return self._menuBar
        menuBar = window.menuBar()
        if not isinstance(menuBar, MenuBar):
            menuBar = MenuBar(window)
            window.setMenuBar(menuBar)
        return menuBar

    def setupMenuBar(self, menuBar=None):
        if menuBar is None:
            try:
                menuBar = self._menuBar
            except Exception:
                return
            menuBar.resetState()
        activeWindow = self.activeWindow()
        fileMenu = menuBar.fetchMenu(Entries.File)
        # HACK: scripting window has its own New/Open;
        # figure out how to do this without explicit blacklist.
        if not isinstance(activeWindow, ScriptingWindow):
            fileMenu.fetchAction(Entries.File_New, self.newFile)
            fileMenu.fetchAction(Entries.File_Open, self.openFile)
        recentFilesMenu = fileMenu.fetchMenu(Entries.File_Open_Recent)
        self.updateRecentFiles(recentFilesMenu)
        if not platformSpecific.mergeOpenAndImport():
            fileMenu.fetchAction(Entries.File_Import, self.importFile)
        fileMenu.fetchAction(Entries.File_Save_All, self.saveAll)
        fileMenu.fetchAction(Entries.File_Exit, self.closeAll)

        editMenu = menuBar.fetchMenu(Entries.Edit)
        editMenu.fetchAction(Entries.Edit_Settings, self.settings)

        viewMenu = menuBar.fetchMenu(Entries.View)
        self.updateDrawingAttributes(viewMenu)

        coordinatesSubmenu = viewMenu.fetchMenu(Entries.View_Show_Coordinates)
        self.updateDrawingAttributes(coordinatesSubmenu)

        scriptsMenu = menuBar.fetchMenu(Entries.Scripts)
        self.updateExtensions(scriptsMenu)

        windowMenu = menuBar.fetchMenu(Entries.Window)
        if platformSpecific.windowCommandsInMenu() and activeWindow is not None:
            windowMenu.fetchAction(Entries.Window_Minimize, activeWindow.showMinimized)
            windowMenu.fetchAction(Entries.Window_Minimize_All, self.minimizeAll)
            windowMenu.fetchAction(Entries.Window_Zoom, lambda: self.zoom(activeWindow))
        windowMenu.fetchAction(Entries.Window_Scripting, self.scripting)
        if self.outputWindow is not None:
            windowMenu.fetchAction(Entries.Window_Output, self.output)
        # TODO: add a list of open windows in window menu, check active window
        # maybe add helper function that filters topLevelWidgets into windows
        # bc we need this in a few places

        helpMenu = menuBar.fetchMenu(Entries.Help)
        helpMenu.fetchAction(
            Entries.Help_Documentation,
            lambda: QDesktopServices.openUrl(QUrl("http://trufont.github.io/")),
        )
        helpMenu.fetchAction(
            Entries.Help_Report_An_Issue,
            lambda: QDesktopServices.openUrl(
                QUrl("https://github.com/trufont/trufont/issues/new")
            ),
        )
        helpMenu.addSeparator()
        helpMenu.fetchAction(Entries.Help_About, self.about)

    def updateMenuBar(self):
        window = self.activeWindow()
        if window is not None and hasattr(window, "setupMenu"):
            menuBar = self.fetchMenuBar(window)
            window.setupMenu(menuBar)
            menuBar.setSpawnElementsHint(False)
            self.setupMenuBar(menuBar)
            menuBar.setSpawnElementsHint(True)
        else:
            self.setupMenuBar()

    # ---------
    # Scripting
    # ---------

    def allFonts(self):
        fonts = []
        for widget in self.topLevelWidgets():
            if isinstance(widget, FontWindow):
                font = widget.font_()
                fonts.append(font)
        return fonts

    def currentFont(self):
        # might be None when closing all windows with scripting window open
        if self._currentFontWindow is None:
            return None
        return self._currentFontWindow.font_()

    def currentGlyph(self):
        return self._currentGlyph

    def setCurrentGlyph(self, glyph):
        if glyph == self._currentGlyph:
            return
        self._currentGlyph = glyph
        self.postNotification("currentGlyphChanged")

    def globals(self):
        global_vars = {
            "__builtins__": __builtins__,
            "AllFonts": self.allFonts,
            "CurrentFont": self.currentFont,
            "CurrentGlyph": self.currentGlyph,
            "events": self.dispatcher,
            "registerExtension": self.registerExtension,
            "unregisterExtension": self.unregisterExtension,
            "registerTool": self.registerTool,
            "unregisterTool": self.unregisterTool,
            "qApp": self,
        }
        return global_vars

    # directory getters

    def _getLocalDirectory(self, key, name):
        userPath = settings.value(key, type=str)
        if userPath and os.path.isdir(userPath):
            return userPath

        appDataFolder = QStandardPaths.standardLocations(
            QStandardPaths.AppLocalDataLocation
        )[0]
        subFolder = os.path.normpath(os.path.join(appDataFolder, name))

        if not os.path.exists(subFolder):
            try:
                os.makedirs(subFolder)
            except OSError:
                subFolder = os.path.expanduser("~")

        settings.setValue(key, subFolder)
        return subFolder

    def getExtensionsDirectory(self):
        return self._getLocalDirectory("scripting/extensionsPath", "Extensions")

    def getScriptsDirectory(self):
        return self._getLocalDirectory("scripting/scriptsPath", "Scripts")

    # -------------
    # Drawing tools
    # -------------

    def drawingTools(self):
        return self._drawingTools

    def registerTool(self, tool):
        self._drawingTools.append(tool)
        data = dict(tool=tool)
        self.postNotification("drawingToolRegistered", data)

    def unregisterTool(self, tool):
        self._drawingTools.remove(tool)
        data = dict(tool=tool)
        self.postNotification("drawingToolUnregistered", data)

    # ----------
    # Extensions
    # ----------

    def extensions(self):
        return self._extensions

    def registerExtension(self, extension):
        self._extensions.append(extension)
        self.updateMenuBar()
        data = dict(extension=extension)
        self.postNotification("extensionRegistered", data)

    def unregisterExtension(self, extension):
        self._extensions.remove(extension)
        self.updateMenuBar()
        data = dict(extension=extension)
        self.postNotification("extensionUnregistered", data)

    def updateExtensions(self, menu):
        def getFunc(ext, path):
            # need a stack frame here to return a unique lambda for each run
            return lambda: ext.run(path)

        menu.clear()
        # also clear submenus
        for child in menu.children():
            if isinstance(child, menu.__class__):
                child.setParent(None)
                child.deleteLater()

        for extension in self._extensions:
            addToMenu = extension.addToMenu
            if addToMenu:
                if isinstance(addToMenu, list):
                    parentMenu = menu.addMenu(extension.name or "")
                else:
                    addToMenu = [addToMenu]
                    parentMenu = menu
                for entry in addToMenu:
                    menuName = entry.get("name")
                    menuPath = entry.get("path")
                    shortcut = entry.get("shortcut")
                    parentMenu.addAction(
                        menuName, getFunc(extension, menuPath), shortcut
                    )
        menu.addSeparator()
        menu.addAction(self.tr(Entries.Scripts_Build_Extension), self.extensionBuilder)

    # ----------------
    # Menu Bar entries
    # ----------------

    def newFile(self):
        font = TFont.new()
        window = FontWindow(font)
        window.show()

    def openFile(self, path=None):
        self._openFile(path, importFile=platformSpecific.mergeOpenAndImport())

    def importFile(self):
        self._openFile(openFile=False, importFile=True)

    def _openFile(self, path=None, openFile=True, importFile=False):
        if not path:
            # formats
            fileFormats = []
            supportedFiles = ""
            if openFile:
                packageAsFile = platformSpecific.treatPackageAsFile()
                if packageAsFile:
                    ufoFormat = "*.ufo"
                    tfExtFormat = "*.tfExt"
                else:
                    ufoFormat = "metainfo.plist"
                    tfExtFormat = "info.plist"
                fileFormats.extend(
                    [
                        self.tr("UFO Fonts {}").format("(%s)" % ufoFormat),
                        self.tr("TruFont Extension {}").format("(%s)" % tfExtFormat),
                    ]
                )
                supportedFiles += f"{ufoFormat} {tfExtFormat} "
            if importFile:
                # TODO: systematize this
                fileFormats.extend(
                    [
                        self.tr("OpenType Font file {}").format("(*.otf *.ttf)"),
                        self.tr("Type1 Font file {}").format("(*.pfa *.pfb)"),
                        self.tr("ttx Font file {}").format("(*.ttx)"),
                        self.tr("WOFF Font file {}").format("(*.woff *.woff2)"),
                    ]
                )
                supportedFiles += "*.otf *.pfa *.pfb *.ttf *.ttx *.woff"
            all_supported_types_filter = self.tr("All supported files {}").format(
                "({})".format(supportedFiles.rstrip())
            )
            fileFormats.extend(
                [all_supported_types_filter, self.tr("All files {}").format("(*.*)")]
            )
            # dialog
            importKey = importFile and not openFile
            state = (
                settings.openFileDialogState()
                if not importKey
                else settings.importFileDialogState()
            )
            directory = (
                None
                if state
                else QStandardPaths.standardLocations(QStandardPaths.DocumentsLocation)[
                    0
                ]
            )
            title = self.tr("Open File") if openFile else self.tr("Import File")
            dialog = QFileDialog(
                self.activeWindow(), title, directory, ";;".join(fileFormats)
            )
            if state:
                dialog.restoreState(state)
            dialog.setAcceptMode(QFileDialog.AcceptOpen)
            dialog.setFileMode(QFileDialog.ExistingFile)
            dialog.selectNameFilter(all_supported_types_filter)
            ret = dialog.exec_()
            # save current directory
            # TODO: should open w/o file chooser also update current dir?
            state = dialog.saveState()
            if importKey:
                settings.setImportFileDialogState(directory)
            else:
                settings.setOpenFileDialogState(directory)
            # cancelled?
            if not ret:
                return
            path = dialog.selectedFiles()[0]
        # sanitize
        path = os.path.normpath(path)
        if ".plist" in os.path.basename(path):
            path = os.path.dirname(path)
        ext = os.path.splitext(path)[1]
        if ext == ".ufo":
            self._loadUFO(path)
        elif ext == ".tfExt":
            self._loadExt(path)
        else:
            self._loadBinary(path)

    def _loadBinary(self, path):
        for widget in self.topLevelWidgets():
            if isinstance(widget, FontWindow):
                font = widget.font_()
                if font is not None and font.binaryPath == path:
                    widget.raise_()
                    return
        font = TFont()
        try:
            font.extract(path)
            self._loadFont(font)
        except Exception as e:
            errorReports.showCriticalException(e)
            return
        self.setCurrentFile(font.binaryPath)

    def _loadExt(self, path):
        # TODO: put version check in place
        e = TExtension(path)
        e.install()

    def _loadUFO(self, path):
        for widget in self.topLevelWidgets():
            if isinstance(widget, FontWindow):
                font = widget.font_()
                if font is not None and font.path == path:
                    widget.raise_()
                    return
        try:
            font = TFont(path)
            self._loadFont(font)
        except Exception as e:
            msg = self.tr("There was an issue opening the font at {}.").format(path)
            errorReports.showCriticalException(e, msg)
            return
        self.setCurrentFile(font.path)

    def _loadFont(self, font):
        currentFont = self.currentFont()
        # Open new font in current font window if it contains an unmodified
        # empty font (e.g. after startup).
        if (
            currentFont is not None
            and currentFont.path is None
            and currentFont.binaryPath is None
            and currentFont.dirty is False
        ):
            window = self._currentFontWindow
            window.setFont_(font)
        else:
            window = FontWindow(font)
        window.show()

    def openRecentFile(self):
        fontPath = self.sender().toolTip()
        self.openFile(fontPath)

    def clearRecentFiles(self):
        settings.setRecentFiles([])

    def saveAll(self):
        for widget in self.topLevelWidgets():
            if isinstance(widget, FontWindow):
                widget.saveFile()

    def closeAll(self):
        for widget in self.topLevelWidgets():
            if isinstance(widget, FontWindow):
                widget.close()
        # loop again to see if user kept font windows open
        for widget in self.topLevelWidgets():
            if isinstance(widget, FontWindow):
                return
        self.quit()

    # Edit

    def settings(self):
        if hasattr(self, "_settingsWindow") and self._settingsWindow.isVisible():
            self._settingsWindow.raise_()
        else:
            self._settingsWindow = SettingsWindow()
            self._settingsWindow.show()

    # Scripts

    def extensionBuilder(self):
        # TODO: don't store, spawn window each time instead
        # or have tabs?
        if not hasattr(self, "_extensionBuilderWindow"):
            self._extensionBuilderWindow = ExtensionBuilderWindow()
        if self._extensionBuilderWindow.isVisible():
            self._extensionBuilderWindow.raise_()
        else:
            self._extensionBuilderWindow.show()

    # Window

    def minimizeAll(self):
        for widget in self.topLevelWidgets():
            if widget.isVisible():
                # additional guard, shouldnt be needed
                # if isinstance(widget, (QMenu, QMenuBar)):
                #     continue
                widget.showMinimized()

    def zoom(self, window):
        if window.isMaximized():
            window.showNormal()
        else:
            window.showMaximized()

    def scripting(self):
        # TODO: don't store, spawn window each time instead
        # or have tabs?
        if not hasattr(self, "_scriptingWindow"):
            self._scriptingWindow = ScriptingWindow()
        if self._scriptingWindow.isVisible():
            self._scriptingWindow.raise_()
        else:
            self._scriptingWindow.show()

    def output(self):
        self.outputWindow.setVisible(not self.outputWindow.isVisible())

    # Help

    def about(self):
        AboutDialog(self.activeWindow()).exec_()

    # ------------
    # Recent files
    # ------------

    def setCurrentFile(self, path):
        if path is None:
            return
        path = os.path.abspath(path)
        recentFiles = settings.recentFiles()
        if path in recentFiles:
            recentFiles.remove(path)
        recentFiles.insert(0, path)
        while len(recentFiles) > MAX_RECENT_FILES:
            del recentFiles[-1]
        settings.setRecentFiles(recentFiles)

    def updateRecentFiles(self, menu):
        # bootstrap
        actions = menu.actions()
        for i in range(MAX_RECENT_FILES):
            try:
                action = actions[i]
            except IndexError:
                action = QAction(menu)
                menu.addAction(action)
            action.setVisible(False)
            action.triggered.connect(self.openRecentFile)
        try:
            actions[MAX_RECENT_FILES]
        except IndexError:
            menu.addSeparator()
            action = QAction(menu)
            action.setText(self.tr("Clear Menu"))
            action.triggered.connect(self.clearRecentFiles)
            menu.addAction(action)

        # fill
        actions = menu.actions()
        recentFiles = settings.recentFiles()
        count = min(len(recentFiles), MAX_RECENT_FILES)
        for index, recentFile in enumerate(recentFiles[:count]):
            action = actions[index]
            shortName = os.path.basename(recentFile.rstrip(os.sep))

            action.setText(shortName)
            action.setToolTip(recentFile)
            action.setVisible(True)
        for index in range(count, MAX_RECENT_FILES):
            actions[index].setVisible(False)

        menu.setEnabled(len(recentFiles))
        # TODO: put recent files in dock on OSX
        # import sys
        # if sys.platform == "darwin":
        #     menu.setAsDockMenu()

    # ------------------
    # Drawing attributes
    # ------------------

    def setDrawingAttribute(self):
        sender = self.sender()
        drawingAttributes = settings.drawingAttributes()
        checked = sender.isChecked()
        for attr in sender.data():
            drawingAttributes[attr] = checked
        settings.setDrawingAttributes(drawingAttributes)
        self.postNotification("preferencesChanged")

    def updateDrawingAttributes(self, menu):
        drawingAttributes = settings.drawingAttributes()
        elements = [
            (
                Entries.View_Show_Points,
                ("showGlyphOnCurvePoints", "showGlyphOffCurvePoints"),
                True,
            ),
            (
                Entries.View_Show_Coordinates_When_Selected,
                ("showGlyphCoordinatesWhenSelected",),
                False,
            ),
            (
                Entries.View_Show_Point_Coordinates,
                ("showGlyphPointCoordinates",),
                False,
            ),
            (
                Entries.View_Show_Bezier_Handles_Coordinates,
                ("showGlyphBezierHandlesCoordinates",),
                False,
            ),
            (
                Entries.View_Show_Metrics,
                (
                    "showGlyphMetrics",
                    "showFontVerticalMetrics",
                    "showFontPostscriptBlues",
                ),
                True,
            ),
            (Entries.View_Show_Images, ("showGlyphImage",), True,),
            (
                Entries.View_Show_Guidelines,
                ("showGlyphGuidelines", "showFontGuidelines"),
                True,
            ),
        ]
        for entry, attrs, checked in elements:
            action = menu.fetchAction(entry)
            action.setCheckable(True)
            action.setChecked(drawingAttributes.get(attrs[0], checked))
            action.setData(attrs)
            action.triggered.connect(self.setDrawingAttribute)



from PyQt5.QtCore import Qt

from defconQt.representationFactories.glyphCellFactory import (
    GlyphCellFactoryDrawingController,
    GlyphCellHeaderHeight,
    GlyphCellMinHeightForHeader,
    GlyphCellMinHeightForMetrics,
)
from defconQt.tools import platformSpecific


def TFGlyphCellFactory(
    glyph,
    width,
    height,
    drawMarkColor=True,
    drawTemplate=True,
    drawHeader=None,
    drawMetrics=None,
    pixelRatio=1.0,
):
    if drawHeader is None:
        drawHeader = height >= GlyphCellMinHeightForHeader
    if drawMetrics is None:
        drawMetrics = height >= GlyphCellMinHeightForMetrics
    obj = TFGlyphCellFactoryDrawingController(
        glyph=glyph,
        font=glyph.font,
        width=width,
        height=height,
        drawMarkColor=drawMarkColor,
        drawTemplate=drawTemplate,
        drawHeader=drawHeader,
        drawMetrics=drawMetrics,
        pixelRatio=pixelRatio,
    )
    return obj.getPixmap()


class TFGlyphCellFactoryDrawingController(GlyphCellFactoryDrawingController):
    def __init__(self, *args, **kwargs):
        if "drawTemplate" in kwargs:
            drawTemplate = kwargs.pop("drawTemplate")
        else:
            drawTemplate = True
        super().__init__(*args, **kwargs)
        self.shouldDrawTemplate = drawTemplate

    def drawCellHorizontalMetrics(self, painter, rect):
        if not self.glyph.template:
            super().drawCellHorizontalMetrics(painter, rect)

    def drawCellVerticalMetrics(self, painter, rect):
        if not self.glyph.template:
            super().drawCellVerticalMetrics(painter, rect)

    def drawCellForeground(self, painter, rect):
        if self.shouldDrawTemplate and self.glyph.template:
            painter.save()
            font = platformSpecific.otherUIFont()
            font.setPointSize(0.425 * self.height)
            painter.setFont(font)
            painter.setPen(Qt.lightGray)
            if self.glyph.unicode is not None:
                uni = self.glyph.unicode
            else:
                uni = 0xFFFD
            painter.drawText(
                0,
                0,
                self.width,
                self.height - GlyphCellHeaderHeight,
                Qt.AlignVCenter | Qt.AlignCenter,
                chr(uni),
            )
            painter.restore()



import os
import platform
import subprocess

from PyQt5.Qt import PYQT_VERSION_STR, QT_VERSION_STR
from PyQt5.QtCore import QEvent, QSize, Qt
from PyQt5.QtWidgets import (
    QApplication,
    QDialog,
    QFrame,
    QHBoxLayout,
    QLabel,
    QSizePolicy,
    QStackedWidget,
    QTextBrowser,
    QVBoxLayout,
    QWidget,
)

from trufont import __file__ as modulePath
from trufont import __version__

try:
    PATH = os.path.abspath(os.path.join(modulePath, "../../.."))
    gitShortHash = subprocess.check_output(
        ["git", "rev-parse", "--short", "HEAD"], cwd=PATH, stderr=subprocess.DEVNULL
    ).decode()
    gitShortLog = subprocess.check_output(
        ["git", "shortlog", "-sn"], cwd=PATH, stderr=subprocess.DEVNULL
    ).decode()
    with open(os.path.join(PATH, "COPYRIGHT"), encoding="utf-8") as fd:
        licenseText = fd.read()
    licenseText = "<p>{}</p>".format(
        licenseText.replace("\n\n", "</p><p>").replace("\n", " ")
    )
    with open(os.path.join(PATH, "THANKS"), encoding="utf-8") as fd:
        thanksText = fd.read()
    thanksText = "<p>{}</p>".format(
        thanksText.replace("\n\n", "</p><p>").replace("\n–", "<br>–").replace("\n", " ")
    )
except Exception:
    gitShortHash = gitShortLog = licenseText = thanksText = ""


class AboutDialog(QDialog):
    def __init__(self, parent=None):
        super().__init__(
            parent,
            Qt.MSWindowsFixedSizeDialogHint
            | Qt.WindowTitleHint
            | Qt.WindowSystemMenuHint
            | Qt.WindowCloseButtonHint,
        )
        self.setAttribute(Qt.WA_DeleteOnClose)
        self.setWindowModality(Qt.ApplicationModal)
        self.setWindowTitle(self.tr("About"))

        app = QApplication.instance()
        name = app.applicationName()
        domain = app.organizationDomain()

        iconLabel = QLabel(self)
        iconLabel.setAlignment(Qt.AlignCenter)
        iconLabel.setMaximumWidth(250)
        iconLabel.setMinimumSize(250, 210)
        icon = self.windowIcon()
        size = icon.actualSize(QSize(186, 186))
        iconLabel.setPixmap(icon.pixmap(size))
        titleLabel = QLabel(self)
        titleLabel.setText(
            self.tr(
                "<p style='color: #353535; font-size: 24pt; font-weight: 250'>"
                "TruFont Font Editor</p>"
                "<p style='font-size: 13pt; font-weight: 400'>{} Pristine Wax</p>"
            ).format(__version__)
        )
        textLabel = QLabel(self)
        text = self.tr(
            "<p>{n} is a free and open source font editor and scripting "
            "environment made by the developers of the {n} community.</p>"
            "<p>{n} is built upon the "
            "<a href='http://ts-defcon.readthedocs.org/en/ufo3/' "
            "style='color: #356FDE'>defcon</a> UFO library and exposes a "
            "<a href='http://robofab.com/' style='color: #356FDE'>robofab</a>"
            "-like API for scripting purposes.</p>"
            "<p>Running on Qt {} (PyQt {}).</p>"
            "<p>Version {} {} – Python {}."
        ).format(
            QT_VERSION_STR,
            PYQT_VERSION_STR,
            __version__,
            gitShortHash,
            platform.python_version(),
            n=name,
        )
        if domain:
            text += self.tr(
                "<br>See <a href='http://{d}' style='color: #356FDE'>{d}</a> "
                "for more information.</p>"
            ).format(d=domain)
        else:
            text += "</p>"
        textLabel.setText(text)
        textLabel.setOpenExternalLinks(True)
        textLabel.setWordWrap(True)

        authorsLabel = QTextBrowser(self)
        authorsLabel.setText("\n".join(self.authors()))

        licenseLabel = QTextBrowser(self)
        licenseLabel.setText(licenseText)

        thanksLabel = QTextBrowser(self)
        thanksLabel.setText(thanksText)

        self.stackWidget = QStackedWidget(self)
        self.stackWidget.addWidget(textLabel)
        self.stackWidget.addWidget(authorsLabel)
        self.stackWidget.addWidget(licenseLabel)
        self.stackWidget.addWidget(thanksLabel)

        textLayout = QVBoxLayout()
        textLayout.setContentsMargins(0, 14, 16, 16)
        textLayout.addWidget(titleLabel)
        textLayout.addWidget(self.stackWidget)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Preferred, QSizePolicy.Expanding)
        textLayout.addWidget(spacer)

        mainLayout = QHBoxLayout()
        mainLayout.addWidget(iconLabel)
        mainLayout.addLayout(textLayout)

        frame = QFrame()
        frame.setMinimumHeight(54)
        frame.setMaximumHeight(54)
        frame.setStyleSheet("background: rgb(230, 230, 230)")

        buttonsLayout = QHBoxLayout(frame)
        for index, text in enumerate(("Authors", "License", "Credits")):
            label = QLabel(text, self)
            label.setAlignment(Qt.AlignCenter)
            label.setCursor(Qt.PointingHandCursor)
            label.setProperty("index", index + 1)
            label.setStyleSheet("color: #356FDE; text-decoration: underline")
            label.installEventFilter(self)
            buttonsLayout.addWidget(label)

        layout = QVBoxLayout(self)
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addLayout(mainLayout)
        layout.addWidget(frame)

    def authors(self):
        for line in gitShortLog.splitlines():
            elem = line.split("\t")[1]
            if not elem or elem.startswith("=?") or elem.endswith("bot"):
                continue
            yield elem

    # ----------
    # Qt methods
    # ----------

    def eventFilter(self, obj, event):
        if event.type() == QEvent.MouseButtonPress:
            index = obj.property("index")
            if index is not None:
                if self.stackWidget.currentIndex() == index:
                    index = 0
                self.stackWidget.setCurrentIndex(index)
                return True
            return False
        return super().eventFilter(obj, event)

    def sizeHint(self):
        return QSize(600, 314)



import os

from fontTools.svgLib import SVGPath
from fontTools.ufoLib.glifLib import readGlyphFromString
from PyQt5.QtCore import (
    QBuffer,
    QByteArray,
    QEvent,
    QIODevice,
    QObject,
    QRectF,
    Qt,
    pyqtSignal,
)
from PyQt5.QtGui import (
    QContextMenuEvent,
    QImage,
    QImageReader,
    QMouseEvent,
    QPainterPath,
    QPainterPathStroker,
    QTransform,
)
from PyQt5.QtWidgets import QApplication

from defconQt.controls.glyphContextView import GlyphContextView, GlyphFlags
from defconQt.controls.glyphView import GlyphViewMinSizeForDetails
from trufont.drawingTools.baseTool import BaseTool
from trufont.objects import settings
from trufont.objects.layoutManager import LayoutManager
from trufont.tools import drawing, errorReports
from trufont.tools.uiMethods import UIGlyphGuidelines

GlyphViewMinSizeForGrid = 10000


class KeyEventFilter(QObject):
    def eventFilter(self, object, event):
        if event.type() == QEvent.ShortcutOverride:
            # we'll only kang shortcut that do not have modifiers
            if event.modifiers() == Qt.NoModifier:
                event.accept()
                return True
        return False


class GlyphCanvasView(GlyphContextView):
    glyphNamesChanged = pyqtSignal()
    toolModified = pyqtSignal(object)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFocusPolicy(Qt.StrongFocus)
        self.setAcceptDrops(True)
        self._currentTool = BaseTool()
        self._currentToolActivated = False
        self._eventFilter = KeyEventFilter(self)
        font = self.window().font_()
        self._layoutManager = LayoutManager(font, self)
        self._mouseDown = False
        self._preview = False

        self._defaultDrawingAttributes["showGlyphSelection"] = True

        # inbound notifications
        app = QApplication.instance()
        app.dispatcher.addObserver(self, "_needsUpdate", "glyphViewUpdate")
        app.dispatcher.addObserver(self, "_preferencesChanged", "preferencesChanged")

        self.readSettings()

    def readSettings(self):
        drawingAttributes = settings.drawingAttributes()
        for attr, value in drawingAttributes.items():
            self.setDefaultDrawingAttribute(attr, value)

    # --------------
    # Custom Methods
    # --------------

    def inputNames(self):
        return self._layoutManager.glyphList()

    def setInputNames(self, names):
        self._layoutManager.setGlyphList(names)

    def layoutManager(self):
        return self._layoutManager

    def setGlyphRecords(self, glyphRecords):
        # TODO: should we also NAK in mouseDown here?
        app = QApplication.instance()
        app.postNotification("glyphViewGlyphsWillChange")
        # TODO: secondly stop calling tool disabled when it's just
        # the glyph changing?
        # e.g. def toolGlyphWillChange(self, newGlyph)
        self._setCurrentToolEnabled(False)
        super().setGlyphRecords(glyphRecords)
        self._setCurrentToolEnabled(True)
        app.postNotification("glyphViewGlyphsChanged")
        self.glyphNamesChanged.emit()

    def previewEnabled(self):
        return self._preview

    def setPreviewEnabled(self, value):
        if value and self._mouseDown:
            return
        if value != self._preview:
            self._preview = value
            if value:
                self.setCursor(Qt.OpenHandCursor)
            else:
                self.setCursor(self._currentTool.cursor)
            self.update()

    # -------------
    # Notifications
    # -------------

    def _subscribeToGlyphs(self, glyphRecords):
        super()._subscribeToGlyphs(glyphRecords)
        handledGlyphs = set()
        for glyphRecord in glyphRecords:
            glyph = glyphRecord.glyph
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.addObserver(self, "_glyphNameChanged", "Glyph.NameChanged")
            glyph.addObserver(self, "_glyphSelectionChanged", "Glyph.SelectionChanged")

    def _unsubscribeFromGlyphs(self):
        handledGlyphs = set()
        for glyphRecord in self._glyphRecords:
            glyph = glyphRecord.glyph
            if glyph in handledGlyphs:
                continue
            handledGlyphs.add(glyph)
            glyph.removeObserver(self, "Glyph.NameChanged")
            glyph.removeObserver(self, "Glyph.SelectionChanged")
        super()._unsubscribeFromGlyphs()

    def _glyphNameChanged(self, notification):
        self.glyphNamesChanged.emit()

    def _glyphSelectionChanged(self, notification):
        self.update()

    # app

    def _needsUpdate(self, notification):
        self.update()

    def _preferencesChanged(self, notification):
        self.readSettings()

    # ---------------
    # Display Control
    # ---------------

    def drawingAttribute(self, attr, flags):
        toolOverride = self._currentTool.drawingAttribute(attr, flags)
        if toolOverride is not None:
            return toolOverride
        return super().drawingAttribute(attr, flags)

    def drawingColor(self, attr, flags):
        toolOverride = self._currentTool.drawingColor(attr, flags)
        if toolOverride is not None:
            return toolOverride
        return super().drawingColor(attr, flags)

    # defaults

    def showSelection(self):
        return self.defaultDrawingAttribute("showGlyphSelection")

    def setShowSelection(self, value):
        self.setDefaultDrawingAttribute("showGlyphSelection", value)

    # ---------------
    # Drawing helpers
    # ---------------

    def drawGlyphBackground(self, painter, glyph, flags):
        if not self._preview:
            super().drawGlyphBackground(painter, glyph, flags)

    def drawBackground(self, painter, index):
        app = QApplication.instance()
        data = dict(widget=self, painter=painter, index=index)
        app.postNotification("glyphViewDrawBackground", data)
        self._currentTool.paintBackground(painter, index)

    def drawGlyphLayer(self, painter, glyph, flags):
        if self._preview:
            if flags.isActiveLayer:
                self.drawFillAndPoints(painter, glyph, flags)
        else:
            super().drawGlyphLayer(painter, glyph, flags)

    def drawForeground(self, painter, index):
        app = QApplication.instance()
        data = dict(widget=self, painter=painter, index=index)
        app.postNotification("glyphViewDrawForeground", data)
        self._currentTool.paint(painter, index)

    # drawing primitives

    def drawMetrics(self, painter, glyph, flags):
        # TODO: should this have its own parameter?
        if self._impliedPointSize > GlyphViewMinSizeForGrid:
            viewportRect = (
                self.mapRectToCanvas(self.rect()).adjusted(0, 0, 2, 2).getRect()
            )
            drawing.drawGrid(painter, self._inverseScale, viewportRect)
        super().drawMetrics(painter, glyph, flags)

    def drawImage(self, painter, glyph, flags):
        drawing.drawGlyphImage(painter, glyph, self._inverseScale)

    def drawGuidelines(self, painter, glyph, flags):
        drawText = self._impliedPointSize > GlyphViewMinSizeForDetails
        viewportRect = self.mapRectToCanvas(self.rect()).getRect()
        if self.drawingAttribute("showFontGuidelines", flags):
            drawing.drawFontGuidelines(
                painter, glyph, self._inverseScale, viewportRect, drawText=drawText
            )
        if self.drawingAttribute("showGlyphGuidelines", flags):
            drawing.drawGlyphGuidelines(
                painter, glyph, self._inverseScale, viewportRect, drawText=drawText
            )

    def drawFillAndPoints(self, painter, glyph, flags):
        contourFillColor = self.drawingColor("contourFillColor", flags)
        componentFillColor = (
            contourFillColor
            if self._preview
            else self.drawingColor("componentFillColor", flags)
        )
        drawFill = self._preview or self.drawingAttribute("showGlyphFill", flags)
        drawComponentFill = self._preview or self.drawingAttribute(
            "showGlyphComponentFill", flags
        )
        drawSelection = not self._preview and self.drawingAttribute(
            "showGlyphSelection", flags
        )
        drawing.drawGlyphFillAndStroke(
            painter,
            glyph,
            self._inverseScale,
            contourFillColor=contourFillColor,
            componentFillColor=componentFillColor,
            drawFill=drawFill,
            drawComponentFill=drawComponentFill,
            drawSelection=drawSelection,
            drawStroke=False,
        )
        if self._preview or not self._impliedPointSize > GlyphViewMinSizeForDetails:
            return
        drawStartPoints = self.drawingAttribute("showGlyphStartPoints", flags)
        drawOnCurves = self.drawingAttribute("showGlyphOnCurvePoints", flags)
        drawOffCurves = self.drawingAttribute("showGlyphOffCurvePoints", flags)
        drawCoordinates = self.drawingAttribute("showGlyphPointCoordinates", flags)
        drawHandleCoordinates = self.drawingAttribute(
            "showGlyphBezierHandlesCoordinates", flags
        )
        drawCoordinatesOnSelection = self.drawingAttribute(
            "showGlyphCoordinatesWhenSelected", flags
        )
        drawing.drawGlyphPoints(
            painter,
            glyph,
            self._inverseScale,
            drawOnCurves=drawOnCurves,
            drawOffCurves=drawOffCurves,
            drawStartPoints=drawStartPoints,
            drawCoordinates=drawCoordinates,
            drawHandleCoordinates=drawHandleCoordinates,
            drawCoordinatesOnSelection=drawCoordinatesOnSelection,
            backgroundColor=self._backgroundColor,
        )

    def drawStroke(self, painter, glyph, flags):
        drawDetails = self._impliedPointSize > GlyphViewMinSizeForDetails
        drawStroke = self.drawingAttribute("showGlyphStroke", flags)
        drawComponentStroke = self.drawingAttribute("showGlyphComponentStroke", flags)
        drawing.drawGlyphFillAndStroke(
            painter,
            glyph,
            self._inverseScale,
            drawFill=False,
            drawComponentFill=False,
            drawStroke=drawStroke,
            drawComponentStroke=drawComponentStroke,
            drawSelection=False,
            partialAliasing=drawDetails,
        )

    def drawAnchors(self, painter, glyph, flags):
        if not self._impliedPointSize > GlyphViewMinSizeForDetails:
            return
        drawing.drawGlyphAnchors(painter, glyph, self._inverseScale)

    # ---------------
    # QWidget methods
    # ---------------

    def showEvent(self, event):
        super().showEvent(event)
        self._setCurrentToolEnabled(True)

    def hideEvent(self, event):
        super().hideEvent(event)
        self._setCurrentToolEnabled(False)
        self._preview = False

    def closeEvent(self, event):
        super().closeEvent(event)
        if event.isAccepted():
            self._setCurrentToolEnabled(False)
            app = QApplication.instance()
            app.dispatcher.removeObserver(self, "glyphViewUpdate")
            app.dispatcher.removeObserver(self, "parametersChanged")

    def dragEnterEvent(self, event):
        mimeData = event.mimeData()
        if mimeData.hasUrls():
            url = mimeData.urls()[0]
            if url.isLocalFile():
                path = url.toLocalFile()
                ext = os.path.splitext(path)[1][1:]
                formats = QImageReader.supportedImageFormats() + ["glif"]
                if ext.lower() in formats:
                    event.acceptProposedAction()
            return
        super().dragEnterEvent(event)

    def dropEvent(self, event):
        mimeData = event.mimeData()
        if mimeData.hasUrls():
            paths = mimeData.urls()
            # pick just one image
            path = paths[0].toLocalFile()
            fileName = os.path.basename(path)
            with open(path, "rb") as imgFile:
                data = imgFile.read()
            ext = os.path.splitext(path)[1][1:]
            # TODO: make sure we cleanup properly when replacing an image with
            # another
            if ext.lower() == "glif":
                otherGlyph = self._glyph.__class__()
                try:
                    readGlyphFromString(data, otherGlyph, otherGlyph.getPointPen())
                except Exception as e:
                    errorReports.showCriticalException(e)
                    return
                self._glyph.beginUndoGroup()
                otherGlyph.drawPoints(self._glyph.getPointPen())
                self._glyph.endUndoGroup()
                return
            if ext.lower() == "svg":
                try:
                    svgPath = SVGPath.fromstring(data)
                except Exception as e:
                    errorReports.showCriticalException(e)
                    return
                self._glyph.beginUndoGroup()
                svgPath.draw(self._glyph.getPen())
                self._glyph.endUndoGroup()
                return
            if ext.lower() != "png":
                # convert
                img = QImage(path)
                data = QByteArray()
                buffer = QBuffer(data)
                buffer.open(QIODevice.WriteOnly)
                img.save(buffer, "PNG")
                # format
                data = bytearray(data)
                fileName = "%s.png" % os.path.splitext(fileName)[0]
            imageSet = self._glyph.font.images
            try:
                imageSet[fileName] = data
            except Exception as e:
                errorReports.showCriticalException(e)
                return
            image = self._glyph.instantiateImage()
            image.fileName = fileName
            self._glyph.image = image
            event.setAccepted(True)
        else:
            super().dropEvent(event)

    def contextMenuEvent(self, event):
        self._redirectEvent(event, self._currentTool.contextMenuEvent, True)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        # TODO: sending an event that doesn't contain the menu isn't
        # extremely useful
        app.postNotification("glyphViewContextMenu", data)

    def keyPressEvent(self, event):
        # TODO: put this in event filter?
        if self._currentTool.grabKeyboard and event.key() == Qt.Key_Escape:
            ok = self.setCurrentTool(self._previousTool)
            if ok:
                self.toolModified.emit(self._previousTool)
        # Note: not needed, only for parity with keyReleaseEvent
        if not self._currentTool.grabKeyboard and event.key() == Qt.Key_Space:
            event.ignore()
            return
        self._redirectEvent(event, self._currentTool.keyPressEvent)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("glyphViewKeyPress", data)

    def keyReleaseEvent(self, event):
        # TODO: I don't know why we have to do this for releaseEvent
        if not self._currentTool.grabKeyboard and event.key() == Qt.Key_Space:
            event.ignore()
            return
        self._redirectEvent(event, self._currentTool.keyReleaseEvent)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("glyphViewKeyRelease", data)

    def mousePressEvent(self, event):
        self._mouseDown = True
        if self._preview:
            self._panOrigin = event.globalPos()
            self.setCursor(Qt.ClosedHandCursor)
            return
        self._redirectEvent(event, self._currentTool.mousePressEvent, True)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("glyphViewMousePress", data)

    def mouseMoveEvent(self, event):
        if self._preview and hasattr(self, "_panOrigin"):
            pos = event.globalPos()
            self.scrollBy(pos - self._panOrigin)
            self._panOrigin = pos
            return
        self._redirectEvent(event, self._currentTool.mouseMoveEvent, True)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("glyphViewMouseMove", data)

    def mouseReleaseEvent(self, event):
        self._redirectEvent(event, self._currentTool.mouseReleaseEvent, True)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("glyphViewMouseRelease", data)
        if hasattr(self, "_panOrigin"):
            if self._preview:
                self.setCursor(Qt.OpenHandCursor)
            del self._panOrigin
        self._mouseDown = False

    def mouseDoubleClickEvent(self, event):
        self._redirectEvent(event, self._currentTool.mouseDoubleClickEvent, True)
        app = QApplication.instance()
        data = dict(event=event, widget=self)
        app.postNotification("glyphViewMouseDoubleClick", data)

    # ------------
    # Canvas tools
    # ------------

    # current tool

    def _setCurrentToolEnabled(self, value):
        if self._currentToolActivated == value:
            return
        self._currentToolActivated = value
        if value:
            self._currentTool.toolActivated()
            if self._currentTool.grabKeyboard:
                self.window().installEventFilter(self._eventFilter)
        else:
            if self._currentTool.grabKeyboard:
                self.window().removeEventFilter(self._eventFilter)
            self._currentTool.toolDisabled()

    def currentTool(self):
        return self._currentTool

    def setCurrentTool(self, tool):
        if self._mouseDown:
            return False
        self._setCurrentToolEnabled(False)
        if tool.grabKeyboard:
            self._previousTool = self._currentTool
        self._currentTool = tool
        self.setCursor(tool.cursor)
        self._setCurrentToolEnabled(True)
        return True

    def _redirectEvent(self, event, callback, transmute=False):
        if self._preview:
            return
        # construct a new event with pos in canvas coordinates
        if transmute:
            if isinstance(event, QContextMenuEvent):
                canvasPos = self.mapToCanvas(event.pos())
                event = event.__class__(
                    event.reason(), canvasPos, event.globalPos(), event.modifiers()
                )
                event.localPos = lambda: canvasPos
            elif isinstance(event, QMouseEvent):
                # TODO: not redirect mouse events if there's no glyph?
                # if not self._glyphRecords:
                #     return
                canvasPos = self.mapToCanvas(event.localPos())
                event = event.__class__(
                    event.type(),
                    canvasPos,
                    event.windowPos(),
                    event.screenPos(),
                    event.button(),
                    event.buttons(),
                    event.modifiers(),
                )
            else:
                raise ValueError(f"cannot transmute event: {event}")
        callback(event)

    # items location

    def _itemsAt(self, func, obj, justOne=True):
        """
        Go through all anchors, points and components (in this order) in the
        glyph, construct their canvas path and list items for which
        *func(path, obj)* returns True, or only return the first item if
        *justOne* is set to True.

        An item is a (contour, index) or (element, None) tuple. Point objects
        aren't returned directly because their usefulness is limited barring
        their parent contour.

        Here is a sample *func* function that tests whether item with path
        *path* contains *pos*:

            def myFunction(path, pos):
                return path.contains(pos)

        This is useful to find out whether an item was clicked on canvas.
        """
        scale = self._inverseScale
        # TODO: export this from drawing or use QSettings.
        # XXX: outdated
        # anchor
        anchorSize = 7 * scale
        anchorHalfSize = anchorSize / 2
        # offCurve
        offSize = 5 * scale
        offHalf = offSize / 2
        offStrokeWidth = 2.5 * scale
        # onCurve
        onSize = 6.5 * scale
        onHalf = onSize / 2
        onStrokeWidth = 1.5 * scale
        # onCurve smooth
        smoothSize = 8 * scale
        smoothHalf = smoothSize / 2
        # guideline pt
        guidelineStrokeWidth = 1 * scale

        if not justOne:
            ret = dict(anchors=[], points=[], components=[], guidelines=[], image=None)
        if self._glyph is None:
            if justOne:
                return None
            return ret
        flags = GlyphFlags(True)
        # anchors
        if self.drawingAttribute("showGlyphAnchors", flags):
            for anchor in reversed(self._glyph.anchors):
                path = QPainterPath()
                path.addEllipse(
                    anchor.x - anchorHalfSize,
                    anchor.y - anchorHalfSize,
                    anchorSize,
                    anchorSize,
                )
                if func(path, obj):
                    if justOne:
                        return anchor
                    ret["anchors"].append(anchor)
        # points
        if self.drawingAttribute("showGlyphOnCurvePoints", flags):
            for contour in reversed(self._glyph):
                for index, point in enumerate(contour):
                    path = QPainterPath()
                    if point.segmentType is None:
                        x = point.x - offHalf
                        y = point.y - offHalf
                        path.addEllipse(x, y, offSize, offSize)
                        strokeWidth = offStrokeWidth
                    else:
                        if point.smooth:
                            x = point.x - smoothHalf
                            y = point.y - smoothHalf
                            path.addEllipse(x, y, smoothSize, smoothSize)
                        else:
                            x = point.x - onHalf
                            y = point.y - onHalf
                            path.addRect(x, y, onSize, onSize)
                        strokeWidth = onStrokeWidth
                    path = _shapeFromPath(path, strokeWidth)
                    if func(path, obj):
                        if justOne:
                            return (contour, index)
                        ret["points"].append((contour, index))
        # components
        for component in reversed(self._glyph.components):
            path = component.getRepresentation("TruFont.QPainterPath")
            if func(path, obj):
                if justOne:
                    return component
                ret["components"].append(component)
        # guideline
        # TODO: we should further dispatch with showFontGuidelines,
        # although both are bind in the UI
        if self.drawingAttribute("showGlyphGuidelines", flags):
            for guideline in UIGlyphGuidelines(self._glyph):
                if None not in (guideline.x, guideline.y):
                    # point
                    x = guideline.x - smoothHalf
                    y = guideline.y - smoothHalf
                    path = QPainterPath()
                    path.addEllipse(x, y, smoothSize, smoothSize)
                    path = _shapeFromPath(path, guidelineStrokeWidth)
                    if func(path, obj):
                        if justOne:
                            return guideline
                        ret["guidelines"].append(guideline)
                    # TODO: catch line if selected
        # image
        if self.drawingAttribute("showGlyphImage", flags):
            image = self._glyph.image
            pixmap = image.getRepresentation("defconQt.QPixmap")
            if pixmap is not None:
                path = QPainterPath()
                transform = QTransform(*image.transformation)
                rect = transform.mapRect(QRectF(pixmap.rect()))
                path.addRect(*rect.getCoords())
                if func(path, obj):
                    if justOne:
                        return image
                    ret["image"] = image
        if not justOne:
            return ret
        return None

    def itemAt(self, pos):
        """
        Find one item at *pos*.

        An item is a (point, contour) or (anchor, None) or (component, None)
        tuple.
        """
        return self.itemsAt(pos, True)

    def itemsAt(self, pos, items=False):
        """
        Find items at *pos*.
        """
        return self._itemsAt(lambda path, pos: path.contains(pos), pos, items)

    def items(self, rect):
        """
        Find items that intersect with *rect* (can be any QPainterPath).
        """
        return self._itemsAt(lambda path, rect: path.intersects(rect), rect, False)


def _shapeFromPath(path, width):
    if path.isEmpty():
        return path
    ps = QPainterPathStroker()
    ps.setWidth(width)
    p = ps.createStroke(path)
    p.addPath(path)
    return p



from PyQt5.QtCore import QSize, Qt
from PyQt5.QtGui import QColor
from PyQt5.QtWidgets import (
    QAbstractItemView,
    QApplication,
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFileDialog,
    QGridLayout,
    QHBoxLayout,
    QLabel,
    QLineEdit,
    QListWidget,
    QListWidgetItem,
    QMenu,
    QPlainTextEdit,
    QPushButton,
    QSizePolicy,
    QSplitter,
    QVBoxLayout,
    QWidget,
)

from defconQt.controls.listView import ListView
from trufont.controls.nameTabWidget import NameTabWidget
from trufont.objects import icons, settings


class SettingsWindow(QDialog):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowTitle(self.tr("Settings"))

        self.tabWidget = NameTabWidget(self)
        self.tabWidget.addNamedTab(GlyphSetTab(self))
        self.tabWidget.addNamedTab(MetricsWindowTab(self))
        self.tabWidget.addNamedTab(MiscTab(self))

        buttonBox = QDialogButtonBox(QDialogButtonBox.Ok | QDialogButtonBox.Cancel)
        buttonBox.accepted.connect(self.accept)
        buttonBox.rejected.connect(self.reject)

        mainLayout = QVBoxLayout()
        mainLayout.addWidget(self.tabWidget)
        mainLayout.addWidget(buttonBox)
        self.setLayout(mainLayout)

        self.readSettings()

    def readSettings(self):
        geometry = settings.settingsWindowGeometry()
        if geometry:
            self.restoreGeometry(geometry)

    def writeSettings(self):
        settings.setSettingsWindowGeometry(self.saveGeometry())

    def moveEvent(self, event):
        self.writeSettings()

    resizeEvent = moveEvent

    def sizeHint(self):
        return QSize(625, 450)

    def accept(self):
        for i in range(self.tabWidget.count()):
            self.tabWidget.widget(i).writeSettings()
        app = QApplication.instance()
        app.postNotification("preferencesChanged")
        super().accept()


class GlyphSetTab(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.name = self.tr("Glyph sets")

        self.defaultGlyphSetBox = QCheckBox(self.tr("Default glyph set:"), self)
        self.defaultGlyphSetDrop = QComboBox(self)
        self.defaultGlyphSetBox.toggled.connect(self.toggleGlyphSetDrop)

        self.glyphSetList = QListWidget(self)
        self.glyphSetList.setSortingEnabled(True)
        self.glyphSetContents = QPlainTextEdit(self)
        self.glyphSetList.currentItemChanged.connect(self.updateGlyphSetContents)
        self.glyphSetList.itemChanged.connect(self.renameGlyphSet)
        self._cachedName = None
        splitter = QSplitter(self)
        splitter.addWidget(self.glyphSetList)
        splitter.addWidget(self.glyphSetContents)
        self.addGlyphSetButton = QPushButton(self)
        self.addGlyphSetButton.setIcon(icons.i_plus())
        self.addGlyphSetButton.clicked.connect(lambda: self.addGlyphSet())
        self.removeGlyphSetButton = QPushButton(self)
        self.removeGlyphSetButton.setIcon(icons.i_minus())
        self.removeGlyphSetButton.clicked.connect(self.removeGlyphSet)
        self.importButton = QPushButton(self.tr("Import"), self)
        importMenu = QMenu(self)
        importMenu.addAction(
            self.tr("Import from Current Font"), self.importFromCurrentFont
        )
        self.importButton.setMenu(importMenu)
        self.glyphListBox = QCheckBox(self.tr("Glyph list path:"), self)
        self.glyphListEdit = QLineEdit(self)
        self.glyphListEdit.setReadOnly(True)
        self.glyphListButton = QPushButton(self.tr("Browse…"), self)
        self.glyphListButton.clicked.connect(self.getGlyphList)
        self.glyphListBox.toggled.connect(self.glyphListEdit.setEnabled)
        self.glyphListBox.toggled.connect(self.glyphListButton.setEnabled)

        buttonsLayout = QHBoxLayout()
        buttonsLayout.addWidget(self.addGlyphSetButton)
        buttonsLayout.addWidget(self.removeGlyphSetButton)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        buttonsLayout.addWidget(spacer)
        buttonsLayout.addWidget(self.importButton)

        firstLayout = QGridLayout()
        line = 0
        firstLayout.addWidget(self.defaultGlyphSetBox, line, 0, 1, 2)
        firstLayout.addWidget(self.defaultGlyphSetDrop, line, 3, 1, 3)
        line += 1
        firstLayout.addWidget(splitter, line, 0, 1, 6)
        line += 1
        firstLayout.addLayout(buttonsLayout, line, 0, 1, 3)
        secondLayout = QHBoxLayout()
        secondLayout.addWidget(self.glyphListBox)
        secondLayout.addWidget(self.glyphListEdit)
        secondLayout.addWidget(self.glyphListButton)
        mainLayout = QVBoxLayout(self)
        mainLayout.addLayout(firstLayout)
        mainLayout.addLayout(secondLayout)
        self.setLayout(mainLayout)

        self.readSettings()

    def addGlyphSet(self, glyphNames=[], glyphSetName=None):
        if glyphSetName is None:
            glyphSetName = self.tr("New glyph set")
        if glyphSetName in self.glyphSets:
            index = 1
            while "%s %d" % (glyphSetName, index) in self.glyphSets:
                index += 1
            glyphSetName = "%s %d" % (glyphSetName, index)
        self.glyphSets[glyphSetName] = glyphNames
        item = QListWidgetItem(glyphSetName, self.glyphSetList)
        item.setFlags(item.flags() | Qt.ItemIsEditable)
        self.glyphSetList.setCurrentItem(item)
        self.glyphSetList.editItem(item)
        self.removeGlyphSetButton.setEnabled(True)

    def removeGlyphSet(self):
        i = self.glyphSetList.currentRow()
        text = self.glyphSetList.takeItem(i).text()
        del self.glyphSets[text]
        if self.glyphSetList.count() < 2:
            self.removeGlyphSetButton.setEnabled(False)

    def renameGlyphSet(self):
        newKey = self.glyphSetList.currentItem()
        if newKey is None:
            return
        newKey = newKey.text()
        self.glyphSets[newKey] = self.glyphSets[self._cachedName]
        del self.glyphSets[self._cachedName]

    def importFromCurrentFont(self):
        font = QApplication.instance().currentFont()
        name = f"{font.info.familyName} {font.info.styleName}"
        self.addGlyphSet(font.glyphOrder, name)

    def toggleGlyphSetDrop(self):
        sender = self.sender()
        self.defaultGlyphSetDrop.setEnabled(sender.isChecked())

    def updateGlyphSetContents(self, current, previous):
        # store content of the textEdit in the glyphSet dict
        if previous is not None:
            glyphNames = self.glyphSetContents.toPlainText().split()
            self.glyphSets[previous.text()] = glyphNames
        # now update the text edit to current glyphSet
        glyphSetName = current.text()
        text = " ".join(self.glyphSets[glyphSetName])
        self.glyphSetContents.setPlainText(text)
        # cache current name for renames
        self._cachedName = glyphSetName

    def getGlyphList(self):
        fileFormats = (
            self.tr("Text file {}").format("(*.txt)"),
            self.tr("All files {}").format("(*.*)"),
        )
        path, _ = QFileDialog.getOpenFileName(
            self, self.tr("Open File"), "", ";;".join(fileFormats)
        )
        if path:
            self.glyphListEdit.setText(path)

    def readSettings(self):
        defaultGlyphSet = settings.defaultGlyphSet()
        self.defaultGlyphSetBox.setChecked(len(defaultGlyphSet))

        self.glyphSets = settings.readGlyphSets()
        self.defaultGlyphSetDrop.clear()
        self.defaultGlyphSetDrop.addItems(self.glyphSets.keys())

        self.glyphSetList.clear()
        glyphSetNames = self.glyphSets.keys()
        # Normally we should be enforcing this rather decently in the interface
        # already
        if glyphSetNames:
            for glyphSetName in glyphSetNames:
                item = QListWidgetItem(glyphSetName, self.glyphSetList)
                item.setFlags(item.flags() | Qt.ItemIsEditable)
            self.glyphSetList.setCurrentRow(0)
        self.removeGlyphSetButton.setEnabled(len(self.glyphSets) > 1)

        glyphListPath = settings.glyphListPath()
        self.glyphListBox.setChecked(bool(glyphListPath))
        self.glyphListEdit.setEnabled(bool(glyphListPath))
        self.glyphListEdit.setText(glyphListPath)
        self.glyphListButton.setEnabled(bool(glyphListPath))

    def writeSettings(self):
        # store content of the textEdit in the glyphSet dict
        glyphNames = self.glyphSetContents.toPlainText().split()
        currentGlyphSet = self.glyphSetList.currentItem().text()
        self.glyphSets[currentGlyphSet] = glyphNames

        settings.writeGlyphSets(self.glyphSets)
        if not self.defaultGlyphSetBox.isChecked():
            settings.setDefaultGlyphSet(None)
        else:
            defaultGlyphSet = self.defaultGlyphSetDrop.currentText()
            settings.setDefaultGlyphSet(defaultGlyphSet)
        if not self.glyphListBox.isChecked():
            settings.setGlyphListPath(None)
        else:
            glyphListPath = self.glyphListEdit.text()
            if glyphListPath:
                settings.setGlyphListPath(glyphListPath)
                QApplication.instance().loadGlyphList()


class MetricsWindowTab(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.name = self.tr("Metrics Window")

        self.inputTextLabel = QLabel(self.tr("Default text:"), self)
        self.inputTextList = QListWidget(self)
        self.inputTextList.setDragDropMode(QAbstractItemView.InternalMove)
        self.addItemButton = QPushButton(self)
        self.addItemButton.setIcon(icons.i_plus())
        self.addItemButton.clicked.connect(self.addItem)
        self.removeItemButton = QPushButton(self)
        self.removeItemButton.setIcon(icons.i_minus())
        self.removeItemButton.clicked.connect(self.removeItem)

        buttonsLayout = QHBoxLayout()
        buttonsLayout.addWidget(self.addItemButton)
        buttonsLayout.addWidget(self.removeItemButton)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        buttonsLayout.addWidget(spacer)

        layout = QVBoxLayout(self)
        layout.addWidget(self.inputTextLabel)
        layout.addWidget(self.inputTextList)
        layout.addLayout(buttonsLayout)
        self.setLayout(layout)

        self.readSettings()

    def addItem(self):
        item = QListWidgetItem(self.inputTextList)
        item.setFlags(item.flags() | Qt.ItemIsEditable)
        self.inputTextList.setCurrentItem(item)
        self.inputTextList.editItem(item)
        self.removeItemButton.setEnabled(True)

    def removeItem(self):
        i = self.inputTextList.currentRow()
        self.inputTextList.takeItem(i)
        if not self.inputTextList.count():
            self.removeItemButton.setEnabled(False)

    def readSettings(self):
        self.inputTextList.clear()
        entries = settings.metricsWindowComboBoxItems()
        for entry in entries:
            item = QListWidgetItem(entry, self.inputTextList)
            item.setFlags(item.flags() | Qt.ItemIsEditable)
        if not len(entries):
            self.removeItemButton.setEnabled(False)

    def writeSettings(self):
        entries = []
        for i in range(self.inputTextList.count()):
            item = self.inputTextList.item(i)
            entries.append(item.text())
        settings.setMetricsWindowComboBoxItems(entries)


class MiscTab(QWidget):
    def __init__(self, parent=None):
        super().__init__(parent)
        self.name = self.tr("Misc")

        self.markColorLabel = QLabel(self.tr("Default flag colors:"), self)
        self.markColorView = ListView(self)
        self.markColorView.setDragEnabled(True)
        # HACK: we need a model before declaring headers
        self.markColorView.setList([])
        self.markColorView.setHeaderLabels((self.tr("Color"), self.tr("Name")))
        self.addItemButton = QPushButton(self)
        self.addItemButton.setIcon(icons.i_plus())
        self.addItemButton.clicked.connect(self.addItem)
        self.removeItemButton = QPushButton(self)
        self.removeItemButton.setIcon(icons.i_minus())
        self.removeItemButton.clicked.connect(self.removeItem)

        self.loadRecentFileBox = QCheckBox(
            self.tr("Load most recent file on start"), self
        )

        buttonsLayout = QHBoxLayout()
        buttonsLayout.setSizeConstraint(QHBoxLayout.SetMinimumSize)
        buttonsLayout.addWidget(self.addItemButton)
        buttonsLayout.addWidget(self.removeItemButton)
        spacer = QWidget()
        spacer.setSizePolicy(QSizePolicy.Expanding, QSizePolicy.Preferred)
        buttonsLayout.addWidget(spacer)

        layout = QVBoxLayout(self)
        layout.addWidget(self.markColorLabel)
        layout.addWidget(self.markColorView)
        layout.addLayout(buttonsLayout)
        layout.addWidget(self.loadRecentFileBox)
        self.setLayout(layout)

        self.readSettings()

    def addItem(self):
        lst = self.markColorView.list()
        lst.append([QColor(170, 255, 255), self.tr("New!")])
        self.markColorView.setList(lst)
        self.markColorView.editIndex(len(lst) - 1, 1)
        self.removeItemButton.setEnabled(True)

    def removeItem(self):
        self.markColorView.removeCurrentRow()
        if not len(self.markColorView.list()):
            self.removeItemButton.setEnabled(False)

    def readSettings(self):
        entries = settings.readMarkColors()
        self.markColorView.setList(entries)
        if not len(entries):
            self.removeItemButton.setEnabled(False)

        loadRecentFile = settings.loadRecentFile()
        self.loadRecentFileBox.setChecked(loadRecentFile)

    def writeSettings(self):
        markColors = self.markColorView.list()
        settings.writeMarkColors(markColors)

        loadRecentFile = self.loadRecentFileBox.isChecked()
        settings.setLoadRecentFile(loadRecentFile)



import re

from PyQt5.QtWidgets import QApplication

_parseGL_RE = re.compile("([A-Za-z_0-9.]+);([0-9A-F]{4})")


def parseGlyphList(path):
    GL2UV = {}
    with open(path) as file:
        for line in file:
            if not line or line[:1] == "#":
                continue
            m = _parseGL_RE.match(line)
            if not m:
                raise SyntaxError(
                    QApplication.translate(
                        "parseGlyphList", "syntax error in glyphlist: {}"
                    ).format(repr(line[:20]))
                )
            glyphName = m.group(1)
            if glyphName in GL2UV:
                print(
                    QApplication.translate(
                        "parseGlyphList",
                        "warning: glyphName redefined in glyphList: {}",
                    ).format(glyphName)
                )
            GL2UV[glyphName] = int(m.group(2), 16)
    return GL2UV



"""
The *baseCodeEditor* submodule
------------------------------

The *baseCodeEditor* submodule provides language-agnostic code editor parts,
including a search widget, goto dialog and code highlighter.
"""

import re

from PyQt5.QtCore import QRegularExpression, QSize, Qt, pyqtSignal
from PyQt5.QtGui import (
    QColor,
    QFontMetricsF,
    QPainter,
    QPalette,
    QRegularExpressionValidator,
    QSyntaxHighlighter,
    QTextCursor,
)
from PyQt5.QtWidgets import (
    QDialog,
    QLabel,
    QLineEdit,
    QPlainTextEdit,
    QVBoxLayout,
    QWidget,
)

from defconQt.tools import drawing, platformSpecific

__all__ = ["GotoLineDialog", "BaseCodeHighlighter", "BaseCodeEditor"]

# -------------------
# Search/Jump dialogs
# -------------------


# TODO: search widget


class GotoLineDialog(QDialog):
    """
    A QDialog_ that asks for a line:column number to the user. Column number is
    optional.

    The result may be passed to the :func:`scrollToLine` function of
    :class:`BaseCodeEditor`.

    .. _QDialog: http://doc.qt.io/qt-5/qdialog.html
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setWindowModality(Qt.WindowModal)
        self.setWindowTitle(self.tr("Go to…"))

        self.lineEdit = QLineEdit(self)
        validator = QRegularExpressionValidator(self)
        validator.setRegularExpression(
            QRegularExpression("(^[1-9][0-9]*(:[1-9][0-9]*)?$)?")
        )
        self.lineEdit.setValidator(validator)
        self.lineEdit.returnPressed.connect(self.accept)
        label = QLabel(self.tr("Enter a row:column to go to"), self)

        layout = QVBoxLayout(self)
        layout.addWidget(self.lineEdit)
        layout.addWidget(label)
        self.setLayout(layout)

    @classmethod
    def getLineColumnNumber(cls, parent):
        dialog = cls(parent)
        result = dialog.exec_()
        newLine = dialog.lineEdit.text()
        if newLine:
            newLine = [int(nb) for nb in newLine.split(":")]
        else:
            newLine = [None]
        if len(newLine) < 2:
            newLine.append(None)
        newLine.append(result)
        return tuple(newLine)


# ------------
# Line numbers
# ------------


class LineNumberArea(QWidget):
    def __init__(self, editor):
        super().__init__(editor)

    def sizeHint(self):
        return QSize(self.parent().lineNumberAreaWidth(), 0)

    def paintEvent(self, event):
        self.parent().lineNumberAreaPaintEvent(event)


# ----------------
# Syntax highlight
# ----------------


class BaseCodeHighlighter(QSyntaxHighlighter):
    """
    A QSyntaxHighlighter_ that highlights code using QRegularExpression_
    (perl regexes).
    Use :func:`addRule` to add a formatting rule or :func:`addBlockRule` to add a
    rule with start and end patterns (e.g., multiline rules).

    # TODO: get/remove rules?

    .. _QRegularExpression: http://doc.qt.io/qt-5/qregularexpression.html
    .. _QSyntaxHighlighter: http://doc.qt.io/qt-5/qsyntaxhighlighter.html
    """

    def __init__(self, parent=None):
        super().__init__(parent)
        self._blockHighlightingRules = []
        self._highlightingRules = []
        self._visited = set()

    def addBlockRule(self, startPattern, endPattern, textFormat):
        """
        Add a block syntax highlighting rule, that begins with regex string
        *startPattern* and ends with *endPattern* formatted according to
        QTextCharFormat_ *format*.

        .. _QTextCharFormat: http://doc.qt.io/qt-5/qtextcharformat.html
        """
        startRegex = QRegularExpression(startPattern)
        endRegex = QRegularExpression(endPattern)
        self._blockHighlightingRules.append((startRegex, endRegex, textFormat))

    def addRule(self, pattern, textFormat):
        """
        Add a syntax highlighting rule with regex string *pattern* and
        QTextCharFormat_ *format*.

        .. _QTextCharFormat: http://doc.qt.io/qt-5/qtextcharformat.html
        """
        regex = QRegularExpression(pattern)
        self._highlightingRules.append((regex, textFormat))

    def highlightBlock(self, text):
        self._visited = set()
        # reset state, in case an earlier block just toggled
        self.setCurrentBlockState(-1)
        # block patterns
        for ident, rules in enumerate(self._blockHighlightingRules):
            startRegex, endRegex, textFormat = rules

            startIndex = offset = 0
            if self.previousBlockState() != ident:
                match = startRegex.match(text)
                startIndex = match.capturedStart()
                offset = match.capturedLength()

            while startIndex >= 0:
                match = endRegex.match(text, startIndex + offset)
                endIndex = match.capturedStart()
                if endIndex == -1:
                    self.setCurrentBlockState(ident)
                    commentLength = len(text) - startIndex
                else:
                    commentLength = endIndex - startIndex + match.capturedLength()
                self.setFormat(startIndex, commentLength, textFormat)
                match = startRegex.match(text, startIndex + commentLength)
                startIndex = match.capturedStart()
                offset = match.capturedLength()

        # inline patterns
        for regex, textFormat in self._highlightingRules:
            it = regex.globalMatch(text)
            while it.hasNext():
                match = it.next()
                start = match.capturedStart()
                length = match.capturedLength()
                self.setFormat(start, length, textFormat)

    def setFormat(self, start, count, *args):
        overlap = False
        for s, e in self._visited:
            if start <= e and s <= start + count:
                overlap = True
        self._visited.add((start, start + count))
        if not overlap:
            super().setFormat(start, count, *args)


# -------------------
# Whitespace Guessing
# -------------------

_whitespaceRE = re.compile("([ \t]+)")


def _guessMinWhitespace(text):
    # gather all whitespace at the beginning of a line
    whitespace = set()
    for line in text.splitlines():
        # skip completely blank lines
        if not line.strip():
            continue
        # store the found whitespace
        m = _whitespaceRE.match(line)
        if m is not None:
            whitespace.add(m.group(1))
    # if nothing was found, fallback to four spaces
    if not whitespace:
        return "    "
    # get the smallest whitespace increment
    whitespace = min(whitespace)
    # if the whitespace starts with a tab, use a single tab
    if whitespace.startswith("\t"):
        return "\t"
    # use what was found
    return whitespace


class BaseCodeEditor(QPlainTextEdit):
    """
    A language-agnostic, abstract code editor. Displays line numbers and
    supports arbitrary indent pattern (Tab/Alt+Tab will add/remove indent,
    Return keeps indent).

    - *indent*: the basic indent pattern. 4-spaces by default.
    - *lineNumbersVisible*: whether line numbers are displayed. True by
      default.
    - *openBlockDelimiter*: a string that indicates a new block. If set, this
      will add an additional level of indent if Return key is pressed after
      such string. None by default.
    - *shouldGuessWhiteSpace*: whether whitespace should be inferred when
      :func:`setPlainText` is called.
    """

    openBlockDelimiter = None
    indentChanged = pyqtSignal(str)

    def __init__(self, parent=None):
        super().__init__(parent)
        self.setFont(platformSpecific.fixedFont())
        self.setTabWidth(4)
        self._indent = "    "
        self._lineNumbersVisible = True
        self._shouldGuessWhitespace = True

        self.lineNumbers = LineNumberArea(self)
        # kick-in geometry update before arming signals
        self.updateLineNumberAreaWidth()
        self.blockCountChanged.connect(self.updateLineNumberAreaWidth)
        self.updateRequest.connect(self.updateLineNumberArea)

    # --------------
    # Custom methods
    # --------------

    def indent(self):
        """
        Returns this widget’s indent pattern.
        """
        return self._indent

    def setIndent(self, indent):
        """
        Sets this widget’s atomic indent pattern to the string *indent*.

        The default is four spaces.

        TODO: reindent document?
        """
        if self._indent == indent:
            return
        self._indent = indent
        self.indentChanged.emit(self._indent)

    def tabWidth(self):
        return self._tabWidth

    def setTabWidth(self, value):
        self._tabWidth = value
        self.updateTabWidth()

    def updateTabWidth(self):
        if not hasattr(self, "_tabWidth"):
            return
        fM = QFontMetricsF(self.font())
        pixelWidth = fM.width(" ") * self._tabWidth
        document = self.document()
        opt = document.defaultTextOption()
        opt.setTabStop(pixelWidth)
        document.setDefaultTextOption(opt)

    def lineNumbersVisible(self):
        """
        Returns whether line numbers are displayed.
        """
        return self._lineNumbersVisible

    def setLineNumbersVisible(self, value):
        """
        Sets whether line numbers should be displayed on the left margin.

        The default is true.
        """
        self._lineNumbersVisible = value
        self.lineNumbers.setVisible(value)
        self.updateLineNumberAreaWidth()

    def shouldGuessWhitespace(self):
        """
        Returns whether this widget infers whitespace when text is set.
        """
        return self._shouldGuessWhitespace

    def setShouldGuessWhitespace(self, value):
        """
        Sets whether this widget should infer whitespace when
        :func:`setPlainText` is called.

        The default is true.
        """
        self._shouldGuessWhitespace = value

    def scrollToLine(self, lineNumber, columnNumber=None):
        """
        Scrolls this widget’s viewport to the line *lineNumber* and sets the
        text cursor to that line, at *columnNumber*. If *columnNumber* is None,
        bookkeeping will be performed.

        Strictly positive numbers are expected.
        """
        lineNumber -= 1
        if columnNumber is None:
            columnNumber = self.textCursor().positionInBlock()
        else:
            columnNumber -= 1
        scrollingUp = lineNumber < self.textCursor().blockNumber()
        # scroll to block
        textBlock = self.document().findBlockByLineNumber(lineNumber)
        newCursor = QTextCursor(textBlock)
        self.setTextCursor(newCursor)
        # make some headroom
        one, two = QTextCursor.Down, QTextCursor.Up
        if scrollingUp:
            one, two = two, one
        for move in (one, one, two, two):
            self.moveCursor(move)
        # address column
        newCursor.movePosition(QTextCursor.NextCharacter, n=columnNumber)
        self.setTextCursor(newCursor)

    # ------------
    # Line numbers
    # ------------

    def lineNumberAreaPaintEvent(self, event):
        painter = QPainter()
        painter.begin(self.lineNumbers)
        rect = event.rect()
        painter.fillRect(rect, self.palette().color(QPalette.Base))
        d = rect.topRight()
        a = rect.bottomRight()
        painter.setPen(QColor(150, 150, 150))
        # Alright. Since we just did setPen() w the default color constructor
        # it set the pen to a cosmetic width (of zero), i.e. drawing one pixel
        # regardless of screen density.
        # The thing is, we want to paint the last pixels at width boundary.
        # So if devicePixelRatio is e.g. 3 (3 device pixels per screen pixel),
        # then we need to translate by 3-1 pixels, but since 3 device pixels
        # is just one pixel is painter units, we'd have to divide by device
        # pixels after, so we would translate by 2/3 "software" pixels.
        pixelRatio = self.lineNumbers.devicePixelRatio()
        delta = (pixelRatio - 1) / pixelRatio
        drawing.drawLine(painter, d.x() + delta, d.y(), a.x() + delta, a.y())
        painter.setFont(self.font())

        block = self.firstVisibleBlock()
        blockNumber = block.blockNumber()
        top = int(
            self.blockBoundingGeometry(block).translated(self.contentOffset()).top()
        )
        bottom = top + int(self.blockBoundingRect(block).height())

        while block.isValid() and top <= event.rect().bottom():
            if block.isVisible() and bottom >= event.rect().top():
                number = str(blockNumber + 1)
                painter.drawText(
                    4,
                    top,
                    self.lineNumbers.width() - 8,
                    self.fontMetrics().height(),
                    Qt.AlignRight,
                    number,
                )
            block = block.next()
            top = bottom
            bottom = top + int(self.blockBoundingRect(block).height())
            blockNumber += 1
        painter.end()

    def lineNumberAreaWidth(self):
        if not self._lineNumbersVisible:
            return 0
        digits = 1
        top = max(1, self.blockCount())
        while top >= 10:
            top /= 10
            digits += 1
        # Avoid too frequent geometry changes
        if digits < 3:
            digits = 3
        return 10 + self.fontMetrics().width("9") * digits

    def updateLineNumberArea(self, rect, dy):
        if dy:
            self.lineNumbers.scroll(0, dy)
        else:
            self.lineNumbers.update(
                0, rect.y(), self.lineNumbers.width(), rect.height()
            )

        if rect.contains(self.viewport().rect()):
            self.updateLineNumberAreaWidth(0)

    def updateLineNumberAreaWidth(self, newBlockCount=None):
        self.setViewportMargins(self.lineNumberAreaWidth(), 0, 0, 0)

    def resizeEvent(self, event):
        super().resizeEvent(event)
        cr = self.contentsRect()
        self.lineNumbers.setGeometry(
            cr.left(), cr.top(), self.lineNumberAreaWidth(), cr.height()
        )

    # ------------
    # Autocomplete
    # ------------

    def findLineIndentLevel(self, cursor=None):
        if cursor is None:
            cursor = self.textCursor()
        indent = 0
        cursor.select(QTextCursor.LineUnderCursor)
        lineLength = len(cursor.selectedText()) // len(self._indent)
        cursor.movePosition(QTextCursor.StartOfLine)
        while lineLength > 0:
            cursor.movePosition(
                QTextCursor.NextCharacter, QTextCursor.KeepAnchor, len(self._indent)
            )
            if cursor.selectedText() == self._indent:
                indent += 1
            else:
                break
            # Now move the anchor back to the position()
            # shouldn't NoMove work here?
            # cursor.movePosition(QTextCursor.NoMove)
            cursor.setPosition(cursor.position())
            lineLength -= 1
        cursor.movePosition(QTextCursor.EndOfLine)
        return indent

    def performLinewiseIndent(self, cursor, positive=True):
        # do a lisewise indent
        p = cursor.position()
        a = cursor.anchor()
        if a > p:
            a, p = p, a

        cursor.setPosition(p)
        pBlock = cursor.blockNumber()
        cursor.setPosition(a)
        aBlock = cursor.blockNumber()

        cursor.beginEditBlock()
        for _ in range(pBlock - aBlock + 1):
            cursor.movePosition(QTextCursor.StartOfBlock)
            if positive:
                cursor.insertText(self._indent)
            else:
                cursor.movePosition(
                    QTextCursor.NextCharacter, QTextCursor.KeepAnchor, len(self._indent)
                )
                if cursor.selectedText() == self._indent:
                    cursor.removeSelectedText()
            cursor.movePosition(QTextCursor.NextBlock)
        cursor.endEditBlock()

    def keyPressEvent(self, event):
        key = event.key()
        if key == Qt.Key_Return:
            cursor = self.textCursor()
            indentLvl = self.findLineIndentLevel()
            if self.openBlockDelimiter is not None:
                cursor.movePosition(
                    QTextCursor.PreviousCharacter, QTextCursor.KeepAnchor
                )
                if cursor.selectedText() == self.openBlockDelimiter:
                    indentLvl += 1
            super().keyPressEvent(event)
            newLineSpace = "".join(self._indent for _ in range(indentLvl))
            cursor = self.textCursor()
            cursor.insertText(newLineSpace)
        elif key in (Qt.Key_Backspace, Qt.Key_Backtab):
            cursor = self.textCursor()
            if key == Qt.Key_Backtab and cursor.hasSelection():
                self.performLinewiseIndent(cursor, False)
            else:
                cursor.movePosition(
                    QTextCursor.PreviousCharacter,
                    QTextCursor.KeepAnchor,
                    len(self._indent),
                )
                if cursor.selectedText() == self._indent:
                    cursor.removeSelectedText()
                else:
                    super().keyPressEvent(event)
        elif key == Qt.Key_Tab:
            cursor = self.textCursor()
            if cursor.hasSelection():
                self.performLinewiseIndent(cursor)
            else:
                cursor.insertText(self._indent)
        else:
            super().keyPressEvent(event)

    # --------------
    # Other builtins
    # --------------

    def setFont(self, font):
        super().setFont(font)
        self.updateTabWidth()

    def setPlainText(self, text):
        super().setPlainText(text)
        if self._shouldGuessWhitespace and text is not None:
            indent = _guessMinWhitespace(text)
            self.setIndent(indent)

    def wheelEvent(self, event):
        if event.modifiers() & platformSpecific.scaleModifier():
            font = self.font()
            newPointSize = font.pointSize() + event.angleDelta().y() / 120.0
            if newPointSize < 6:
                return
            font.setPointSize(newPointSize)
            self.setFont(font)
        else:
            super().wheelEvent(event)



from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import QLabel


# A QLabel for right-aligning text to edit boxes to cut down repetition.
class RLabel(QLabel):
    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.setAlignment(Qt.AlignRight | Qt.AlignVCenter)
